# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import csv
from abc import abstractmethod
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, List, NoReturn, Optional, Tuple, TypeVar

import numpy as np
import torch
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.algorithms.moo.rnsga2 import RNSGA2
from pymoo.core.problem import Problem
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.repair.rounding import RoundingRepair
from pymoo.operators.sampling.rnd import IntegerRandomSampling
from pymoo.optimize import minimize
from torch.utils.data.dataloader import DataLoader

import nncf
from nncf import NNCFConfig
from nncf.common.initialization.batchnorm_adaptation import BatchnormAdaptationAlgorithm
from nncf.common.logging import nncf_logger
from nncf.common.plotting import noninteractive_plotting
from nncf.common.utils.decorators import skip_if_dependency_unavailable
from nncf.common.utils.os import safe_open
from nncf.common.utils.registry import Registry
from nncf.config.extractors import get_bn_adapt_algo_kwargs
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_controller import ElasticityController
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.multi_elasticity_handler import SubnetConfig
from nncf.experimental.torch.nas.bootstrapNAS.search.evaluator import AccuracyEvaluator
from nncf.experimental.torch.nas.bootstrapNAS.search.evaluator import BaseEvaluator
from nncf.experimental.torch.nas.bootstrapNAS.search.evaluator import MACsEvaluator
from nncf.experimental.torch.nas.bootstrapNAS.search.evaluator_handler import AccuracyEvaluatorHandler
from nncf.experimental.torch.nas.bootstrapNAS.search.evaluator_handler import BaseEvaluatorHandler
from nncf.experimental.torch.nas.bootstrapNAS.search.evaluator_handler import EfficiencyEvaluatorHandler
from nncf.torch.nncf_network import NNCFNetwork

SEARCH_ALGORITHMS = Registry("search algorithm", add_name_as_attr=True)

DataLoaderType = TypeVar("DataLoaderType")
TModel = TypeVar("TModel")
ValFnType = Callable[[TModel, DataLoaderType], float]


class FixIntegerRandomSampling(IntegerRandomSampling):
    """
    Wrapper for the IntegerRandomSampling with the fix for https://github.com/anyoptimization/pymoo/issues/388.
    """

    def _do(self, problem, n_samples, **kwargs):
        n, (xl, xu) = problem.n_var, problem.bounds()
        return np.column_stack([np.random.randint(xl[k], xu[k] + 1, size=(n_samples)) for k in range(n)])


class EvolutionaryAlgorithms(Enum):
    NSGA2 = "NSGA2"
    RNSGA2 = "RNSGA2"


class SearchParams:
    """
    Storage class for search parameters.
    """

    def __init__(
        self,
        num_evals: float,
        num_constraints: float,
        population: float,
        seed: float,
        crossover_prob: float,
        crossover_eta: float,
        mutation_prob: float,
        mutation_eta: float,
        acc_delta: float,
        ref_acc: float,
    ):
        """
        Initializes storage class for search parameters.

        :param num_evals: Number of evaluations for the search algorithm.
        :param num_constraints: Number of constraints in search problem
        :param population: Population size
        :param seed: Seed used by the search algorithm.
        :param crossover_prob: Crossover probability used by evolutionary algorithm
        :param crossover_eta: Crossover eta
        :param mutation_prob: Mutation probability
        :param mutation_eta: Mutation eta
        :param acc_delta: Tolerated accuracy delta to select a single sub-network from Pareto front.
        :param ref_acc: Accuracy of input model or reference.
        """
        self.num_constraints = num_constraints
        self.population = population
        if population > num_evals:
            raise ValueError("Population size must not be greater than number of evaluations.")
        self.num_evals = num_evals // population * population
        self.seed = seed
        self.crossover_prob = crossover_prob
        self.crossover_eta = crossover_eta
        self.mutation_prob = mutation_prob
        self.mutation_eta = mutation_eta
        self.acc_delta = acc_delta
        self.ref_acc = ref_acc

    @classmethod
    def from_dict(cls, search_config: Dict[str, Any]) -> "SearchParams":
        """
        Initializes search params storage class from Dict.

        :param search_config: Dictionary with search configuration.
        :return: Instance of the storage class
        """
        num_evals = search_config.get("num_evals", 3000)
        num_constraints = search_config.get("num_constraints", 0)
        population = search_config.get("population", 40)
        seed = search_config.get("seed", 0)
        crossover_prob = search_config.get("crossover_prob", 0.9)
        crossover_eta = search_config.get("crossover_eta", 10.0)
        mutation_prob = search_config.get("mutation_prob", 0.02)
        mutation_eta = search_config.get("mutation_eta", 3.0)
        acc_delta = search_config.get("acc_delta", 1)
        ref_acc = search_config.get("ref_acc", -1)

        return cls(
            num_evals,
            num_constraints,
            population,
            seed,
            crossover_prob,
            crossover_eta,
            mutation_prob,
            mutation_eta,
            acc_delta,
            ref_acc,
        )


class RNSGA2SearchParams(SearchParams):
    """
    Storage class for search parameters of RNSGA-II algorithm.
    """

    def __init__(
        self,
        aspiration_points: np.ndarray,
        epsilon: float,
        weights: np.ndarray,
        extreme_points_as_ref_points: bool,
        **kwargs,
    ):
        """
        Initializes storage class for search parameters of RNSGA-II.

        :param aspiration_points: Aspiration or reference points for RNSGA-II.
        :param epsilon: epsilon distance of surviving solutions for RNSGA-II .
        :param weights: weights used by RNSGA-II.
        :param extreme_points_as_ref_points: Find extreme points and use them as aspiration points.
        """
        super().__init__(**kwargs)
        self.aspiration_points = aspiration_points
        self.epsilon = epsilon
        self.weights = weights
        self.extreme_points_as_ref_points = extreme_points_as_ref_points

    @classmethod
    def from_dict(cls, search_config: Dict[str, Any]) -> "RNSGA2SearchParams":
        """
        Initializes search params storage class from Dict.

        :param search_config: Dictionary with search configuration.
        :return: Instance of the storage class
        """
        num_evals = search_config.get("num_evals", 3000)
        num_constraints = search_config.get("num_constraints", 0)
        population = search_config.get("population", 40)
        seed = search_config.get("seed", 0)
        crossover_prob = search_config.get("crossover_prob", 0.9)
        crossover_eta = search_config.get("crossover_eta", 10.0)
        mutation_prob = search_config.get("mutation_prob", 0.02)
        mutation_eta = search_config.get("mutation_eta", 3.0)
        acc_delta = search_config.get("acc_delta", 1)
        ref_acc = search_config.get("ref_acc", -1)
        aspiration_points = search_config.get("aspiration_points", [[0, 0]])
        aspiration_points = np.array(aspiration_points)
        epsilon = search_config.get("epsilon", 0.001)
        weights = search_config.get("weights")
        extreme_points_as_ref_points = search_config.get("extreme_points_as_ref_points", False)

        return cls(
            aspiration_points,
            epsilon,
            weights,
            extreme_points_as_ref_points,
            num_evals=num_evals,
            num_constraints=num_constraints,
            population=population,
            seed=seed,
            crossover_prob=crossover_prob,
            crossover_eta=crossover_eta,
            mutation_prob=mutation_prob,
            mutation_eta=mutation_eta,
            acc_delta=acc_delta,
            ref_acc=ref_acc,
        )


class BaseSearchAlgorithm:
    """
    Base class for search algorithms. It contains the evaluators used by search approaches.
    """

    def __init__(self):
        """
        Initialize BaseSearchAlgorithm class.

        """
        self._use_default_evaluators = True
        self._evaluator_handlers = []
        self._accuracy_evaluator_handler = None
        self._efficiency_evaluator_handler = None
        self._log_dir = None
        self._search_records = []
        self.bad_requests = []
        self.best_config = None
        self.best_vals = None
        self.best_pair_objective = float("inf")
        self._tb = None

    @classmethod
    def from_config(cls, model, elasticity_ctrl, nncf_config):
        """
        Construct a search algorithm
        :param model: Super-Network
        :param elasticity_ctrl: Interface to manage elasticity of super-network
        :param nncf_config: Dict with configuration for search algorithm
        :return: instance of the search algorithm.
        """
        search_config = nncf_config.get("bootstrapNAS", {}).get("search", {})
        algo_name = search_config.get("algorithm")
        algo_cls = SEARCH_ALGORITHMS.get(algo_name)
        if not algo_name:
            raise NotImplementedError(f"Evolutionary Search Algorithm {algo_name} not implemented")
        return algo_cls(model, elasticity_ctrl, nncf_config)

    @classmethod
    def from_checkpoint(
        cls, model: NNCFNetwork, elasticity_ctrl: ElasticityController, bn_adapt_args, resuming_checkpoint_path: str
    ) -> "BaseSearchAlgorithm":
        raise NotImplementedError("Evolutionary Search Algorithm from checkpoint not implemented")

    @property
    def search_records(self):
        return self._search_records

    @abstractmethod
    def run(
        self,
        validate_fn: Callable,
        val_loader: DataLoader,
        checkpoint_save_dir: str,
        efficiency_evaluator: Optional[BaseEvaluator] = None,
        ref_acc: Optional[float] = 100,
        tensorboard_writer: Optional["SummaryWriter"] = None,  # noqa: F821
    ) -> Tuple[ElasticityController, SubnetConfig, Tuple[float, ...]]:
        """This method should implement how to run the search algorithm."""

    def search_progression_to_csv(self, filename="search_progression.csv") -> NoReturn:
        """
        Exports search progression to CSV file

        :param filename: path to save the CSV file.
        :return:
        """
        with safe_open(Path(self._log_dir) / filename, "w", encoding="utf8") as progression:
            writer = csv.writer(progression)
            for record in self.search_records:
                writer.writerow(record)


@SEARCH_ALGORITHMS.register(EvolutionaryAlgorithms.NSGA2.value)
class NSGA2SearchAlgorithm(BaseSearchAlgorithm):
    def __init__(
        self, model: NNCFNetwork, elasticity_ctrl: ElasticityController, nncf_config: NNCFConfig, verbose=True
    ):
        """
        Initializes search algorithm

        :param model: Super-network
        :param elasticity_ctrl: interface to manage the elasticity of the super-network.
        :param nncf_config: Configuration file.
        :param verbose:
        """
        super().__init__()
        self._model = model
        self._elasticity_ctrl = elasticity_ctrl
        search_config = nncf_config.get("bootstrapNAS", {}).get("search", {})
        self.num_obj = None
        self.search_params = SearchParams.from_dict(search_config)
        self._log_dir = nncf_config.get("log_dir", ".")
        self._verbose = verbose
        self._top1_accuracy_validation_fn = None
        self._val_loader = None
        self._algorithm = NSGA2(
            pop_size=self.search_params.population,
            sampling=FixIntegerRandomSampling(),
            crossover=SBX(
                prob=self.search_params.crossover_prob,
                eta=self.search_params.crossover_eta,
                vtype=float,
                repair=RoundingRepair(),
            ),
            mutation=PM(
                prob=self.search_params.mutation_prob,
                eta=self.search_params.mutation_eta,
                vtype=float,
                repair=RoundingRepair(),
            ),
            eliminate_duplicates=True,
            save_history=False,
        )
        self._num_vars = 0
        self._vars_lower = 0
        self._vars_upper = []

        self._num_vars, self._vars_upper = self._elasticity_ctrl.multi_elasticity_handler.get_design_vars_info()
        if self._num_vars == 0 or self._vars_lower is None:
            raise nncf.InternalError("Search space is empty")

        self._result = None
        bn_adapt_params = search_config.get("batchnorm_adaptation", {})
        bn_adapt_algo_kwargs = get_bn_adapt_algo_kwargs(nncf_config, bn_adapt_params)
        self.bn_adaptation = BatchnormAdaptationAlgorithm(**bn_adapt_algo_kwargs) if bn_adapt_algo_kwargs else None

        self._problem = None
        self.checkpoint_save_dir = None
        self.type_var = int

    @property
    def evaluator_handlers(self) -> List[BaseEvaluatorHandler]:
        """
        Gets a list of the evaluators used by the search algorithm.

        :return: List of available evaluators.
        """
        if self._evaluator_handlers:
            return self._evaluator_handlers
        raise nncf.ValidationError("Evaluator handlers haven't been defined")

    @property
    def acc_delta(self) -> float:
        """
        :return: Value allowed to deviate from when selecting a sub-network in the pareto front.
        """
        return self.search_params.acc_delta

    @acc_delta.setter
    def acc_delta(self, val: float) -> NoReturn:
        """
        :param val: value to update the allowed accuracy delta.
        :return:
        """
        val = min(val, 50)
        if val > 5:
            nncf_logger.warning("Accuracy delta was set to a value greater than 5")
        self.search_params.acc_delta = val

    @property
    def vars_lower(self) -> List[float]:
        """
        Gets access to design variables lower bounds.
        :return: lower bounds for design variables
        """
        return self._vars_lower

    @property
    def vars_upper(self) -> List[float]:
        """
        Gets access to design variables upper bounds.
        :return: upper bounds for design variables
        """
        return self._vars_upper

    @property
    def num_vars(self) -> int:
        """
        Number of design variables used by the search algorithm.
        :return:
        """
        return self._num_vars

    def run(
        self,
        validate_fn: ValFnType,
        val_loader: DataLoaderType,
        checkpoint_save_dir: str,
        efficiency_evaluator: Optional[BaseEvaluator] = None,
        ref_acc: Optional[float] = 100,
        tensorboard_writer: Optional["SummaryWriter"] = None,  # noqa: F821
        evaluator_checkpoint=None,
    ) -> Tuple[ElasticityController, SubnetConfig, Tuple[float, ...]]:
        """
        Runs the search algorithm

        :param validate_fn: Function used to validate the accuracy of the model.
        :param val_loader: Data loader used by the validation function.
        :param checkpoint_save_dir: Path to save checkpoints.
        :param efficiency_evaluator: External efficiency evaluator.
        :param ref_acc: Reference Accuracy for sub-network selection.
        :param tensorboard_writer: Tensorboard writer to log data points.
        :param evaluator_checkpoint:
        :return: Elasticity controller with discovered sub-network, sub-network configuration and
                its performance metrics.
        """
        self._elasticity_ctrl.multi_elasticity_handler.activate_maximum_subnet()
        nncf_logger.info("Searching for optimal subnet.")
        if ref_acc != 100:
            self.search_params.ref_acc = ref_acc
        self._tb = tensorboard_writer
        self.checkpoint_save_dir = checkpoint_save_dir
        self._accuracy_evaluator_handler = AccuracyEvaluatorHandler(
            AccuracyEvaluator(self._model, validate_fn, val_loader), self._elasticity_ctrl
        )
        self._accuracy_evaluator_handler.update_reference_accuracy(self.search_params)
        self.num_obj = 2
        if efficiency_evaluator is not None:
            self._use_default_evaluators = False
            self._efficiency_evaluator_handler = EfficiencyEvaluatorHandler(efficiency_evaluator, self._elasticity_ctrl)
        else:
            self._use_default_evaluators = True

            def get_macs_for_active_subnet() -> float:
                flops, _ = self._elasticity_ctrl.multi_elasticity_handler.count_flops_and_weights_for_active_subnet()
                return flops / 2000000  # MACs

            self._efficiency_evaluator_handler = EfficiencyEvaluatorHandler(
                MACsEvaluator(get_macs_for_active_subnet), self._elasticity_ctrl
            )
        self._evaluator_handlers.append(self._efficiency_evaluator_handler)
        self._evaluator_handlers.append(self._accuracy_evaluator_handler)
        self.maximal_vals = [evaluator_handler.current_value for evaluator_handler in self._evaluator_handlers]

        self._problem = SearchProblem(self)
        self._result = minimize(
            self._problem,
            self._algorithm,
            ("n_gen", int(self.search_params.num_evals / self.search_params.population)),
            seed=self.search_params.seed,
            # save_history=True,
            copy_algorithm=False,
            verbose=self._verbose,
        )

        if self.best_config is not None:
            self._elasticity_ctrl.multi_elasticity_handler.activate_subnet_for_config(self.best_config)
            if self.bn_adaptation is not None:
                self.bn_adaptation.run(self._model)
            ret_vals = self.best_vals
        else:
            nncf_logger.warning("Couldn't find a subnet that satisfies the requirements. Returning maximum subnet.")
            self._elasticity_ctrl.multi_elasticity_handler.activate_maximum_subnet()
            if self.bn_adaptation is not None:
                self.bn_adaptation.run(self._model)
            self.best_config = self._elasticity_ctrl.multi_elasticity_handler.get_active_config()
            self.best_vals = [None, None]
            ret_vals = self.maximal_vals

        return self._elasticity_ctrl, self.best_config, [abs(elem) for elem in ret_vals if elem is not None]

    @skip_if_dependency_unavailable(dependencies=["matplotlib.pyplot"])
    def visualize_search_progression(self, filename="search_progression") -> NoReturn:
        """
        Visualizes search progression and saves the resulting figure.

        :param filename:
        :return:
        """
        import matplotlib.pyplot as plt

        with noninteractive_plotting():
            plt.figure()
            colormap = plt.cm.get_cmap("viridis")
            col = range(int(self.search_params.num_evals / self.search_params.population))
            for i in range(0, len(self.search_records), self.search_params.population):
                c = [col[int(i / self.search_params.population)]] * len(
                    self.search_records[i : i + self.search_params.population]
                )
                plt.scatter(
                    [abs(row[2]) for row in self.search_records][i : i + self.search_params.population],
                    [abs(row[4]) for row in self.search_records][i : i + self.search_params.population],
                    s=9,
                    c=c,
                    alpha=0.5,
                    marker="D",
                    cmap=colormap,
                )
            plt.scatter(
                *tuple(abs(ev.input_model_value) for ev in self.evaluator_handlers),
                marker="s",
                s=120,
                color="blue",
                label="Input Model",
                edgecolors="black",
            )
            if None not in self.best_vals:
                plt.scatter(
                    *tuple(abs(val) for val in self.best_vals),
                    marker="o",
                    s=120,
                    color="yellow",
                    label="BootstrapNAS A",
                    edgecolors="black",
                    linewidth=2.5,
                )
            plt.legend()
            plt.title("Search Progression (NSGA2)")
            plt.xlabel(self.efficiency_evaluator_handler.name)
            plt.ylabel(self.accuracy_evaluator_handler.name)
            plt.savefig(f"{self._log_dir}/{filename}.png")

    def save_evaluators_state(self) -> NoReturn:
        """
        Save state of evaluators used in search.
        :return:
        """
        evaluator_handlers_state = []
        for evaluator_handler in self._evaluator_handlers:
            eval_state = evaluator_handler.evaluator.get_state()
            evaluator_handlers_state.append(eval_state)
        torch.save(evaluator_handlers_state, Path(self.checkpoint_save_dir, "evaluators_state.pth"))

    def evaluators_to_csv(self) -> NoReturn:
        """
        Export evaluators' information used by search algorithm to CSV
        :return:
        """
        for evaluator_handler in self.evaluator_handlers:
            evaluator_handler.export_cache_to_csv(self._log_dir)

    @property
    def efficiency_evaluator_handler(self):
        return self._efficiency_evaluator_handler

    @property
    def accuracy_evaluator_handler(self):
        return self._accuracy_evaluator_handler


@SEARCH_ALGORITHMS.register(EvolutionaryAlgorithms.RNSGA2.value)
class RNSGA2SearchAlgorithm(NSGA2SearchAlgorithm):
    def __init__(
        self, model: NNCFNetwork, elasticity_ctrl: ElasticityController, nncf_config: NNCFConfig, verbose=True
    ):
        """
        Initializes search algorithm

        :param model: Super-network
        :param elasticity_ctrl: interface to manage the elasticity of the super-network.
        :param nncf_config: Configuration file.
        :param verbose:
        """
        super().__init__(model, elasticity_ctrl, nncf_config, verbose)
        search_config = nncf_config.get("bootstrapNAS", {}).get("search", {})
        self.search_params = RNSGA2SearchParams.from_dict(search_config)
        self._algorithm = RNSGA2(
            ref_points=self.search_params.aspiration_points,
            pop_size=self.search_params.population,
            epsilon=self.search_params.epsilon,
            normalization="front",
            extreme_points_as_reference_points=self.search_params.extreme_points_as_ref_points,
            sampling=FixIntegerRandomSampling(),
            crossover=SBX(
                prob=self.search_params.crossover_prob,
                eta=self.search_params.crossover_eta,
                vtype=float,
                repair=RoundingRepair(),
            ),
            mutation=PM(
                prob=self.search_params.mutation_prob,
                eta=self.search_params.mutation_eta,
                vtype=float,
                repair=RoundingRepair(),
            ),
            weights=self.search_params.weights,
            eliminate_duplicates=True,
            save_history=False,
        )

    @skip_if_dependency_unavailable(dependencies=["matplotlib.pyplot"])
    def visualize_search_progression(self, filename="search_progression") -> NoReturn:
        """
        Visualizes search progression and saves the resulting figure.

        :param filename:
        :return:
        """
        import matplotlib.pyplot as plt

        with noninteractive_plotting():
            plt.figure()
            colormap = plt.cm.get_cmap("viridis")
            col = range(int(self.search_params.num_evals / self.search_params.population))
            for i in range(0, len(self.search_records), self.search_params.population):
                c = [col[int(i / self.search_params.population)]] * len(
                    self.search_records[i : i + self.search_params.population]
                )
                plt.scatter(
                    [abs(row[2]) for row in self.search_records][i : i + self.search_params.population],
                    [abs(row[4]) for row in self.search_records][i : i + self.search_params.population],
                    s=9,
                    c=c,
                    alpha=0.5,
                    marker="D",
                    cmap=colormap,
                )
            plt.scatter(
                *tuple(abs(ev.input_model_value) for ev in self.evaluator_handlers),
                marker="s",
                s=120,
                color="blue",
                label="Input Model",
                edgecolors="black",
            )
            if None not in self.best_vals:
                plt.scatter(
                    *tuple(abs(val) for val in self.best_vals),
                    marker="o",
                    s=120,
                    color="yellow",
                    label="BootstrapNAS A",
                    edgecolors="black",
                    linewidth=2.5,
                )
            for point in self._algorithm.survival.ref_points:
                plt.scatter(
                    point[0],
                    point[1],
                    marker="^",
                    color="gray",
                    label="Reference Points",
                    edgecolors="black",
                    linewidth=2.5,
                )
            plt.legend()
            plt.title("Search Progression (RNSGA2)")
            plt.xlabel(self.efficiency_evaluator_handler.name)
            plt.ylabel(self.accuracy_evaluator_handler.name)
            plt.savefig(f"{self._log_dir}/{filename}.png")


class SearchProblem(Problem):
    """
    Pymoo problem with design variables and evaluation methods.
    """

    def __init__(self, search: BaseSearchAlgorithm):
        """
        Initializes search problem

        :param search: search algorithm.
        """
        super().__init__(
            n_var=search.num_vars,
            n_obj=search.num_obj,
            n_constr=search.search_params.num_constraints,
            xl=search.vars_lower,
            xu=search.vars_upper,
            type_var=search.type_var,
        )
        self._search = search
        self._search_records = search.search_records
        self._elasticity_handler = self._search._elasticity_ctrl.multi_elasticity_handler
        self._dims_enabled = self._elasticity_handler.get_available_elasticity_dims()
        self._iter = 0
        self._evaluator_handlers = search.evaluator_handlers
        self._accuracy_evaluator_handler = search.accuracy_evaluator_handler
        self._efficiency_evaluator_handler = search.efficiency_evaluator_handler
        self._model = search._model
        self._lower_bound_acc = search.search_params.ref_acc - search.acc_delta

    def _evaluate(self, x: List[float], out: Dict[str, Any], *args, **kargs) -> NoReturn:
        """
        Evaluates a population of sub-networks.

        :param x: set of sub-networks to evaluate.
        :param out: measurements obtained by evaluating sub-networks.
        :param args:
        :param kargs:
        :return:
        """
        evaluators_arr = [[] for i in range(len(self._search.evaluator_handlers))]

        for _, x_i in enumerate(x):
            sample = self._elasticity_handler.get_config_from_pymoo(x_i)
            self._elasticity_handler.activate_subnet_for_config(sample)
            if sample != self._elasticity_handler.get_active_config():
                nncf_logger.debug("Requested configuration was invalid")
                nncf_logger.debug(f"Requested: {sample}")
                nncf_logger.debug(f"Provided: {self._elasticity_handler.get_active_config()}")
                self._search.bad_requests.append((sample, self._elasticity_handler.get_active_config()))

            result = [sample]

            bn_adaption_executed = False
            for eval_idx, evaluator_handler in enumerate(self._evaluator_handlers):
                in_cache, value = evaluator_handler.retrieve_from_cache(tuple(x_i))
                if not in_cache:
                    if not bn_adaption_executed and self._search.bn_adaptation is not None:
                        self._search.bn_adaptation.run(self._model)
                        bn_adaption_executed = True
                    value = evaluator_handler.evaluate_and_add_to_cache_from_pymoo(tuple(x_i))
                evaluators_arr[eval_idx].append(value)

                result.append(evaluator_handler.name)
                result.append(value)

            self._save_checkpoint_best_subnetwork(sample)
            self._search_records.append(result)

        self._iter += 1
        out["F"] = np.column_stack(list(evaluators_arr))

    def _save_checkpoint_best_subnetwork(self, config: SubnetConfig) -> NoReturn:
        """
        Saves information of current best sub-network discovered by the search algorithm.

        :param config: Best sub-network configuration
        :return:
        """
        acc_within_tolerance = self._accuracy_evaluator_handler.current_value
        pair_objective = self._efficiency_evaluator_handler.current_value
        if acc_within_tolerance < (self._lower_bound_acc * -1.0) and pair_objective < self._search.best_pair_objective:
            self._search.best_pair_objective = pair_objective
            self._search.best_config = config
            self._search.best_vals = [evaluator_handler.current_value for evaluator_handler in self._evaluator_handlers]
            checkpoint_path = Path(self._search.checkpoint_save_dir, "subnetwork_best.pth")
            checkpoint = {
                "best_acc1": acc_within_tolerance * -1.0,
                "best_efficiency": pair_objective,
                "subnet_config": config,
            }
            torch.save(checkpoint, checkpoint_path)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.common.graph.operator_metatypes import InputNoopMetatype
from nncf.common.graph.patterns import GraphPattern
from nncf.common.graph.patterns import HWFusedPatternNames
from nncf.common.utils.registry import Registry
from nncf.onnx.graph.metatypes import onnx_metatypes as om
from nncf.onnx.graph.metatypes.groups import ARITHMETIC_OPERATIONS
from nncf.onnx.graph.metatypes.groups import ATOMIC_ACTIVATIONS_OPERATIONS
from nncf.onnx.graph.metatypes.groups import BATCH_NORMALIZATION_OPERATIONS
from nncf.onnx.graph.metatypes.groups import LINEAR_OPERATIONS

ONNX_HW_FUSED_PATTERNS = Registry("onnx")

# BLOCK PATTERNS


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.MVN)
def create_mvn() -> GraphPattern:
    pattern = GraphPattern()
    pattern_input_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "*INPUT_NODE*", GraphPattern.METATYPE_ATTR: GraphPattern.NON_PATTERN_NODE_TYPE}
    )
    reduce_mean_node_1 = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "REDUCE_MEAN_1", GraphPattern.METATYPE_ATTR: om.ONNXReduceMeanMetatype}
    )
    sub_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "SUBTRACT",
            GraphPattern.METATYPE_ATTR: [om.ONNXSubMetatype],
        }
    )
    pow_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "POW",
            GraphPattern.METATYPE_ATTR: [om.ONNXPowMetatype],
        }
    )
    reduce_mean_node_2 = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "REDUCE_MEAN_2", GraphPattern.METATYPE_ATTR: om.ONNXReduceMeanMetatype}
    )
    add_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD", GraphPattern.METATYPE_ATTR: om.ONNXAddLayerMetatype})
    sqrt_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SQRT", GraphPattern.METATYPE_ATTR: om.ONNXSqrtMetatype})
    div_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "DIV", GraphPattern.METATYPE_ATTR: om.ONNXDivLayerMetatype})

    pattern.add_edge(pattern_input_node, reduce_mean_node_1)
    pattern.add_edge(reduce_mean_node_1, sub_node)
    pattern.add_edge(pattern_input_node, sub_node)
    pattern.add_edge(sub_node, pow_node)
    pattern.add_edge(pow_node, reduce_mean_node_2)
    pattern.add_edge(reduce_mean_node_2, add_node)
    pattern.add_edge(add_node, sqrt_node)
    pattern.add_edge(sqrt_node, div_node)
    pattern.add_edge(sub_node, div_node)
    return pattern


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.MVN_SCALE_SHIFT)
def create_mvn_scale_shift() -> GraphPattern:
    mvn = create_mvn()
    scale_shift = create_scale_shift()

    mvn.join_patterns(scale_shift)
    return mvn


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.GELU)
def create_gelu() -> GraphPattern:
    pattern = GraphPattern()
    pattern_input_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "*INPUT_NODE*", GraphPattern.METATYPE_ATTR: GraphPattern.NON_PATTERN_NODE_TYPE}
    )
    div_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "DIV",
            GraphPattern.METATYPE_ATTR: [om.ONNXDivLayerMetatype, om.ONNXMulLayerMetatype],
        }
    )
    erf_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "ERF",
            GraphPattern.METATYPE_ATTR: om.ONNXErfMetatype,
        }
    )
    add_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "ADD",
            GraphPattern.METATYPE_ATTR: [om.ONNXAddLayerMetatype, om.ONNXSubMetatype],
        }
    )
    mul_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "MUL",
            GraphPattern.METATYPE_ATTR: [om.ONNXMulLayerMetatype, om.ONNXDivLayerMetatype],
        }
    )
    pattern.add_edge(pattern_input_node, div_node)
    pattern.add_edge(div_node, erf_node)
    pattern.add_edge(erf_node, add_node)
    pattern.add_edge(add_node, mul_node)
    pattern.add_edge(pattern_input_node, mul_node)
    return pattern


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.SCALE_SHIFT)
def create_scale_shift() -> GraphPattern:
    pattern = GraphPattern()
    pattern_input_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "*INPUT_NODE*", GraphPattern.METATYPE_ATTR: GraphPattern.NON_PATTERN_NODE_TYPE}
    )
    mul_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.ONNXMulLayerMetatype}
    )
    add_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "ADD, SUBTRACT",
            GraphPattern.METATYPE_ATTR: [om.ONNXAddLayerMetatype, om.ONNXSubMetatype],
        }
    )
    pattern.add_edge(pattern_input_node, mul_node)
    pattern.add_edge(mul_node, add_node)
    return pattern


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.SHIFT_SCALE)
def create_shift_scale() -> GraphPattern:
    pattern = GraphPattern()
    add_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "ADD, SUBTRACT",
            GraphPattern.METATYPE_ATTR: [om.ONNXAddLayerMetatype, om.ONNXSubMetatype],
        }
    )
    mul_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "MULTIPLY, DIV",
            GraphPattern.METATYPE_ATTR: [om.ONNXMulLayerMetatype, om.ONNXDivLayerMetatype],
        }
    )

    pattern.add_edge(add_node, mul_node)
    return pattern


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.SWISH_WITH_SIGMOID)
def create_swish_with_sigmoid() -> GraphPattern:
    pattern = GraphPattern()
    pattern_input_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "*INPUT_NODE*", GraphPattern.METATYPE_ATTR: GraphPattern.NON_PATTERN_NODE_TYPE}
    )
    sigmoid_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "SIGMOID", GraphPattern.METATYPE_ATTR: om.ONNXSigmoidMetatype}
    )
    mul_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.ONNXMulLayerMetatype}
    )

    pattern.add_edge(pattern_input_node, sigmoid_node)
    pattern.add_edge(sigmoid_node, mul_node)
    pattern.add_edge(pattern_input_node, mul_node)
    return pattern


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.SWISH_WITH_HARD_SIGMOID)
def create_swish_with_hard_sigmoid() -> GraphPattern:
    pattern = GraphPattern()
    pattern_input_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "*INPUT_NODE*", GraphPattern.METATYPE_ATTR: GraphPattern.NON_PATTERN_NODE_TYPE}
    )
    hard_sigmoid_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "HARD_SIGMOID", GraphPattern.METATYPE_ATTR: om.ONNXHardSigmoidMetatype}
    )
    mul_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.ONNXMulLayerMetatype}
    )

    pattern.add_edge(pattern_input_node, hard_sigmoid_node)
    pattern.add_edge(hard_sigmoid_node, mul_node)
    pattern.add_edge(pattern_input_node, mul_node)
    return pattern


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.HSWISH_ACTIVATION_WITHOUT_DENOMINATOR)
def create_hswish_without_denominator() -> GraphPattern:
    pattern = GraphPattern()
    any_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "ANY", GraphPattern.METATYPE_ATTR: GraphPattern.ANY_PATTERN_NODE_TYPE}
    )
    add_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD", GraphPattern.METATYPE_ATTR: om.ONNXAddLayerMetatype})
    relu_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "RELU", GraphPattern.METATYPE_ATTR: om.ONNXReluMetatype})
    multiply_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.ONNXMulLayerMetatype}
    )

    pattern.add_edge(any_node, add_node)
    pattern.add_edge(add_node, relu_node)
    pattern.add_edge(relu_node, multiply_node)
    pattern.add_edge(any_node, multiply_node)
    return pattern


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.HSWISH_ACTIVATION)
def create_hswish() -> GraphPattern:
    div_pattern = GraphPattern()
    hswish = create_hswish_without_denominator()
    div_pattern.add_node(**{GraphPattern.LABEL_ATTR: "DIV", GraphPattern.METATYPE_ATTR: om.ONNXDivLayerMetatype})
    hswish.join_patterns(div_pattern)
    return hswish


# INPUT PROCESSING


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_SCALE_SHIFT)
def create_input_scale_shift() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: InputNoopMetatype})
    scale_shift = create_scale_shift()

    pattern.join_patterns(scale_shift)
    return pattern


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_SHIFT_SCALE)
def create_input_shift_scale() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: InputNoopMetatype})
    shift_scale = create_shift_scale()

    pattern.join_patterns(shift_scale)
    return pattern


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_PROCESSING)
def create_input_add() -> GraphPattern:
    pattern = GraphPattern()
    input_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: InputNoopMetatype}
    )
    add_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "ADD, MULTIPLY",
            GraphPattern.METATYPE_ATTR: [om.ONNXAddLayerMetatype, om.ONNXMulLayerMetatype],
        }
    )

    pattern.add_edge(input_node, add_node)
    return pattern


# COMBINATIONS


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ACTIVATIONS_BATCH_NORM)
def create_activations_batch_norm() -> GraphPattern:
    activations = atomic_activations_operations()
    batch_norm = batch_normalization_operations()
    activations.join_patterns(batch_norm)
    return activations


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ACTIVATIONS_SCALE_SHIFT)
def create_activations_scale_shift() -> GraphPattern:
    activations = atomic_activations_operations()
    scale_shift = create_scale_shift()
    activations.join_patterns(scale_shift)
    return activations


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ARITHMETIC_BATCH_NORM)
def create_arithmetic_batch_norm() -> GraphPattern:
    arithmetic = arithmetic_operations()
    batch_norm = batch_normalization_operations()
    arithmetic.join_patterns(batch_norm)
    return arithmetic


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ARITHMETIC_ACTIVATIONS)
def create_arithmetic_activations() -> GraphPattern:
    arithmetic = arithmetic_operations()
    activations = atomic_activations_operations()
    arithmetic.join_patterns(activations)
    return arithmetic


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ARITHMETIC_SCALE_SHIFT)
def create_arithmetic_scale_shift() -> GraphPattern:
    arithmetic = arithmetic_operations()
    scale_shift = create_scale_shift()
    arithmetic.join_patterns(scale_shift)
    return arithmetic


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ARITHMETIC_BATCH_NORM_ACTIVATIONS)
def create_arithmetic_batch_norm_activations() -> GraphPattern:
    arithmetic_batch_norm = create_arithmetic_batch_norm()
    activations = atomic_activations_operations()
    arithmetic_batch_norm.join_patterns(activations)
    return arithmetic_batch_norm


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ARITHMETIC_SCALE_SHIFT_ACTIVATIONS)
def create_arithmetic_scale_shift_activations() -> GraphPattern:
    arithmetic_scale_shift = create_arithmetic_scale_shift()
    activations = atomic_activations_operations()
    arithmetic_scale_shift.join_patterns(activations)
    return arithmetic_scale_shift


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ARITHMETIC_ACTIVATIONS_BATCH_NORM)
def create_arithmetic_activations_batch_norm() -> GraphPattern:
    arithmetic_activations = create_arithmetic_activations()
    batch_norm = batch_normalization_operations()
    arithmetic_activations.join_patterns(batch_norm)
    return arithmetic_activations


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ARITHMETIC_ACTIVATIONS_SCALE_SHIFT)
def create_arithmetic_activations_scale_shift() -> GraphPattern:
    arithmetic_activations = create_arithmetic_activations()
    scale_shift = create_scale_shift()
    arithmetic_activations.join_patterns(scale_shift)
    return arithmetic_activations


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.BATCH_NORM_ACTIVATIONS)
def create_batch_norm_activations() -> GraphPattern:
    batch_norm = batch_normalization_operations()
    activations = atomic_activations_operations()
    batch_norm.join_patterns(activations)
    return batch_norm


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.SCALE_SHIFT_ACTIVATIONS)
def create_scale_shift_activations() -> GraphPattern:
    scale_shift = create_scale_shift()
    activations = atomic_activations_operations()
    scale_shift.join_patterns(activations)
    return scale_shift


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ARITHMETIC)
def create_linear_arithmetic() -> GraphPattern:
    linear = linear_operations()
    arithmetic = arithmetic_operations()
    linear.join_patterns(arithmetic)
    return linear


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_BATCH_NORM)
def create_linear_batch_norm() -> GraphPattern:
    linear = linear_operations()
    batch_norm = batch_normalization_operations()
    linear.join_patterns(batch_norm)
    return linear


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ACTIVATIONS)
def create_linear_activations() -> GraphPattern:
    linear = linear_operations()
    activations = atomic_activations_operations()
    linear.join_patterns(activations)
    return linear


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_BATCH_NORM_ACTIVATIONS)
def create_linear_batch_norm_activations() -> GraphPattern:
    linear_batch_norm = create_linear_batch_norm()
    activations = atomic_activations_operations()
    linear_batch_norm.join_patterns(activations)
    return linear_batch_norm


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_SCALE_SHIFT_ACTIVATIONS)
def create_linear_scale_shift_activations() -> GraphPattern:
    linear_scale_shift = create_linear_scale_shift()
    activations = atomic_activations_operations()
    linear_scale_shift.join_patterns(activations)
    return linear_scale_shift


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ACTIVATIONS_BATCH_NORM)
def create_linear_activations_batch_norm() -> GraphPattern:
    linear_activations = create_linear_activations()
    batch_norm = batch_normalization_operations()
    linear_activations.join_patterns(batch_norm)
    return linear_activations


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ACTIVATIONS_SCALE_SHIFT)
def create_linear_activations_scale_shift() -> GraphPattern:
    linear_activations = create_linear_activations()
    scale_shift = create_scale_shift()
    linear_activations.join_patterns(scale_shift)
    return linear_activations


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_BATCH_NORM_SCALE_SHIFT_ACTIVATIONS)
def create_linear_bn_scale_shift_activation() -> GraphPattern:
    linear_batch_norm = create_linear_batch_norm()
    scale_shift = create_scale_shift()
    activations = atomic_activations_operations()

    linear_batch_norm.join_patterns(scale_shift)
    linear_batch_norm.join_patterns(activations)
    return linear_batch_norm


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_SQUEEZE_ACTIVATIONS)
def create_linear_squeeze_activation() -> GraphPattern:
    linear = linear_operations()
    squeeze = squeeze_operation()
    activations = atomic_activations_operations()

    linear.join_patterns(squeeze)
    linear.join_patterns(activations)
    return linear


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_SQUEEZE_ARITHMETIC_ACTIVATIONS)
def create_linear_squeeze_arithmetic_activation() -> GraphPattern:
    linear = linear_operations()
    squeeze = squeeze_operation()
    arithmetic_activations = create_arithmetic_activations()

    linear.join_patterns(squeeze)
    linear.join_patterns(arithmetic_activations)
    return linear


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.BATCH_NORM_SCALE_SHIFT_ACTIVATIONS)
def create_bn_scale_shift_activation() -> GraphPattern:
    batch_norm = batch_normalization_operations()
    scale_shift = create_scale_shift()
    activations = atomic_activations_operations()

    batch_norm.join_patterns(scale_shift)
    batch_norm.join_patterns(activations)
    return batch_norm


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ARITHMETIC_ACTIVATIONS)
def create_linear_arithmetic_activations() -> GraphPattern:
    linear = linear_operations()
    arithmetic = arithmetic_operations()
    activations = atomic_activations_operations()

    linear.join_patterns(arithmetic)
    linear.join_patterns(activations)
    return linear


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_SHIFT_SCALE)
def create_linear_shift_scale() -> GraphPattern:
    linear = linear_operations()
    shift_scale = create_shift_scale()

    linear.join_patterns(shift_scale)
    return linear


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ARITHMETIC_ACTIVATIONS_ARITHMETIC)
def create_linear_arithmetic_activations_arithmetic() -> GraphPattern:
    linear_arithmetic_activations = create_linear_arithmetic_activations()
    arithmetic = arithmetic_operations()

    linear_arithmetic_activations.join_patterns(arithmetic)
    return linear_arithmetic_activations


# DEVICE PATTERNS


@ONNX_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_SCALE_SHIFT)
def create_linear_scale_shift() -> GraphPattern:
    linear = linear_operations()
    batch_norm = create_scale_shift()
    linear.join_patterns(batch_norm)
    return linear


def linear_operations() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.METATYPE_ATTR: LINEAR_OPERATIONS, GraphPattern.LABEL_ATTR: "LINEAR"})
    return pattern


def batch_normalization_operations() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(
        **{GraphPattern.METATYPE_ATTR: BATCH_NORMALIZATION_OPERATIONS, GraphPattern.LABEL_ATTR: "BATCH_NORMALIZATION"}
    )
    return pattern


def atomic_activations_operations() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(
        **{GraphPattern.METATYPE_ATTR: ATOMIC_ACTIVATIONS_OPERATIONS, GraphPattern.LABEL_ATTR: "ATOMIC_ACTIVATIONS"}
    )

    swish_sigmoid = create_swish_with_sigmoid()
    pattern.add_pattern_alternative(swish_sigmoid)

    swish_hard_sigmoid = create_swish_with_hard_sigmoid()
    pattern.add_pattern_alternative(swish_hard_sigmoid)

    hswish = create_hswish()
    pattern.add_pattern_alternative(hswish)

    hswish_without_denominator = create_hswish_without_denominator()
    pattern.add_pattern_alternative(hswish_without_denominator)

    gelu = create_gelu()
    pattern.add_pattern_alternative(gelu)
    return pattern


def arithmetic_operations() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.METATYPE_ATTR: ARITHMETIC_OPERATIONS, GraphPattern.LABEL_ATTR: "ARITHMETIC"})
    return pattern


def squeeze_operation() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.LABEL_ATTR: "SQUEEZE", GraphPattern.METATYPE_ATTR: om.ONNXSqueezeMetatype})
    return pattern



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import TypeVar

from nncf.common.graph.graph import NNCFGraph
from nncf.common.quantization.quantizer_setup import SingleConfigQuantizerSetup

TModel = TypeVar("TModel")


class Quantizer(ABC):
    """
    Quantizer is an interface for the RangeEstimator algorithm
    which specifies all the required methods to retrieve quantization setup from the given model.
    """

    @abstractmethod
    def transform_prior_quantization(self, model: TModel) -> TModel:
        """
        Transforms the given model in-place with the necessary modifications required prior to quantization.

        :param model: Backend-specific model to be transformed.
        :return: Transformed backend-specific model.
        """

    @abstractmethod
    def get_quantization_setup(self, model: TModel, nncf_graph: NNCFGraph) -> SingleConfigQuantizerSetup:
        """
        Builds SingleConfigQuantizerSetup for the given model.

        :param model: Backend-specific model, for which Quantization Target Points are being seek.
        :param nncf_graph: NNCFGraph instance.
        :return: SingleConfigQuantizerSetup for the given model.
        """



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Tuple

import tensorflow as tf

from nncf.common.utils.registry import Registry


def replace_value_by_index(xs: Tuple[Any, ...], pos: int, value: Any) -> Tuple[Any, ...]:
    """
    Return a new instance of the tuple replacing the specified
    position with the new value.

    :param xs: A tuple.
    :param pos: Zero-based index of the item which should be replaced.
    :param value: New value.
    """
    return tuple(value if idx == pos else elem for idx, elem in enumerate(xs))


def check_port_id(port_id: int, min_port_id: int, max_port_id: int):
    if min_port_id <= port_id <= max_port_id:
        return
    raise ValueError(f"Unexpected `port_id`: {port_id}")


TF_ARG_PROVIDERS = Registry("TF_ARG_PROVIDERS")


class ArgProvider:
    """
    Base class for all argument providers. An `ArgProvider` instance
    describes how to extract input or output argument with specified
    `port_id` from the `args` and `kwargs`.
    """

    def get_input(self, input_port_id: int, args, kwargs) -> tf.Tensor:
        """
        Returns input Tensor with specified `input_port_id`.
        :param input_port_id: Zero-based index an input Tensor
            for TensorFlow operation.
        :return: A Tensor with specified `input_port_id`.
        """

    def get_output(self, output_port_id: int, args, kwargs) -> tf.Tensor:
        """
        Returns output Tensor with specified `output_port_id`.
        :param output_port_id: Zero-based index an output Tensor
            of TensorFlow operation.
        :return: A Tensor with specified `output_port_id`.
        """

    def set_input(self, input_port_id: int, value: tf.Tensor, args, kwargs):
        """
        Updates input Tensor with specified `input_port_id`.
        :param input_port_id: Zero-based index an input Tensor
            for TensorFlow operation which should be updated.
        :return: A tuple (args, kwargs) with updated Tensor.
        """

    def set_output(self, output_port_id: int, value: tf.Tensor, args, kwargs):
        """
        Updates output Tensor with specified `output_port_id`.
        :param output_port_id: Zero-based index an output Tensor
            for TensorFlow operation which should be updated.
        :return: A tuple (args, kwargs) with updated Tensor.
        """


@TF_ARG_PROVIDERS.register("Relu6")
@TF_ARG_PROVIDERS.register("Relu")
@TF_ARG_PROVIDERS.register("Mean")
@TF_ARG_PROVIDERS.register("AddV2")
@TF_ARG_PROVIDERS.register("Placeholder")
@TF_ARG_PROVIDERS.register("BiasAdd")
@TF_ARG_PROVIDERS.register("GatherV2")
@TF_ARG_PROVIDERS.register("Cast")
@TF_ARG_PROVIDERS.register("Sub")
@TF_ARG_PROVIDERS.register("SquaredDifference")
@TF_ARG_PROVIDERS.register("Rsqrt")
@TF_ARG_PROVIDERS.register("Mul")
@TF_ARG_PROVIDERS.register("Erf")
@TF_ARG_PROVIDERS.register("BatchMatMulV2")
@TF_ARG_PROVIDERS.register("RealDiv")
class SimpleOutputArgProvider(ArgProvider):
    def get_output(self, output_port_id: int, args, kwargs) -> tf.Tensor:
        check_port_id(output_port_id, min_port_id=0, max_port_id=0)

        if len(args) > 1:
            raise ValueError(f"Unexpected `args`: {args}")

        return args[output_port_id]

    def set_output(self, output_port_id: int, value: tf.Tensor, args, kwargs):
        check_port_id(output_port_id, min_port_id=0, max_port_id=0)

        if len(args) > 1:
            raise ValueError(f"Unexpected `args`: {args}")

        return replace_value_by_index(args, output_port_id, value), kwargs


@TF_ARG_PROVIDERS.register("Transpose")
class TransposeArgProvider(ArgProvider):
    def get_input(self, input_port_id: int, args, kwargs) -> tf.Tensor:
        check_port_id(input_port_id, min_port_id=0, max_port_id=0)
        return args[input_port_id]

    def set_input(self, input_port_id: int, value: tf.Tensor, args, kwargs):
        check_port_id(input_port_id, min_port_id=0, max_port_id=0)
        return replace_value_by_index(args, input_port_id, value), kwargs


@TF_ARG_PROVIDERS.register("ResizeNearestNeighbor")
class ResizeNearestNeighborArgProvider(ArgProvider):
    """
    Argument provider for the `ResizeNearestNeighbor` operation.
    """

    def get_output(self, output_port_id: int, args, kwargs) -> tf.Tensor:
        check_port_id(output_port_id, min_port_id=0, max_port_id=0)

        if len(args) > 1:
            raise ValueError(f"Unexpected `args`: {args}")

        return args[output_port_id]

    def set_output(self, output_port_id: int, value: tf.Tensor, args, kwargs):
        check_port_id(output_port_id, min_port_id=0, max_port_id=0)

        if len(args) > 1:
            raise ValueError(f"Unexpected `args`: {args}")

        return replace_value_by_index(args, output_port_id, value), kwargs

    def get_input(self, input_port_id: int, args, kwargs) -> tf.Tensor:
        check_port_id(input_port_id, min_port_id=0, max_port_id=0)
        return args[0]

    def set_input(self, input_port_id: int, value: tf.Tensor, args, kwargs):
        check_port_id(input_port_id, min_port_id=0, max_port_id=0)
        return replace_value_by_index(args, input_port_id, value), kwargs


@TF_ARG_PROVIDERS.register("Conv2D")
class Conv2DArgProvider(ArgProvider):
    """
    Argument provider of the `Conv2D` operation.
    Inputs:
        port_id: 0 - input tensor.
        port_id: 1 - filter tensor.
    Outputs:
        port_id: 0 - output tensor.
    """

    def get_input(self, input_port_id: int, args, kwargs) -> tf.Tensor:
        check_port_id(input_port_id, min_port_id=0, max_port_id=1)

        if input_port_id == 0:
            return args[0]

        return kwargs["filter"]  # input_port_id == 1

    def set_input(self, input_port_id: int, value: tf.Tensor, args, kwargs):
        check_port_id(input_port_id, min_port_id=0, max_port_id=1)

        if input_port_id == 0:
            return replace_value_by_index(args, input_port_id, value), kwargs

        kwargs["filter"] = value
        return args, kwargs

    def get_output(self, output_port_id: int, args, kwargs) -> tf.Tensor:
        check_port_id(output_port_id, min_port_id=0, max_port_id=0)

        if len(args) > 1:
            raise ValueError(f"Unexpected `args`: {args}")

        return args[output_port_id]

    def set_output(self, output_port_id: int, value: tf.Tensor, args, kwargs):
        check_port_id(output_port_id, min_port_id=0, max_port_id=0)

        if len(args) > 1:
            raise ValueError(f"Unexpected `args`: {args}")

        return replace_value_by_index(args, output_port_id, value), kwargs


@TF_ARG_PROVIDERS.register("FusedBatchNormV3")
class FusedBatchNormV3ArgProvider(ArgProvider):
    """
    Argument provider of the `FusedBatchNormV3` operation.
    Outputs:
        port_id: 0 - output tensor (`y` key).
    """

    def get_output(self, output_port_id: int, args, kwargs) -> tf.Tensor:
        check_port_id(output_port_id, min_port_id=0, max_port_id=0)
        return args[0].y

    def set_output(self, output_port_id: int, value: tf.Tensor, args, kwargs):
        check_port_id(output_port_id, min_port_id=0, max_port_id=0)
        x = args[0]._replace(y=value)
        return replace_value_by_index(args, 0, x), kwargs


@TF_ARG_PROVIDERS.register("DepthwiseConv2dNative")
class DepthwiseConv2dNativeArgProvider(ArgProvider):
    """
    Argument provider of the `DepthwiseConv2dNative` operation.
    Inputs:
        port_id: 0 - input tensor (args[0]).
        port_id: 1 - filter tensor (args[1]).
    """

    def get_input(self, input_port_id: int, args, kwargs) -> tf.Tensor:
        check_port_id(input_port_id, min_port_id=0, max_port_id=1)
        return args[input_port_id]

    def set_input(self, input_port_id: int, value: tf.Tensor, args, kwargs):
        check_port_id(input_port_id, min_port_id=0, max_port_id=1)
        return replace_value_by_index(args, input_port_id, value), kwargs


@TF_ARG_PROVIDERS.register("MatMul")
class MatMulArgProvider(SimpleOutputArgProvider):
    """
    Argument provider of the `MatMul` operation.
    Inputs:
        port_id: 0 - input tensor (args[0]).
        port_id: 1 - filter (always?) tensor (args[1]).
    """

    def get_input(self, input_port_id: int, args, kwargs) -> tf.Tensor:
        check_port_id(input_port_id, min_port_id=0, max_port_id=1)

        if len(args) == 0:
            return kwargs["a" if input_port_id == 0 else "b"]
        return args[input_port_id]

    def set_input(self, input_port_id: int, value: tf.Tensor, args, kwargs):
        check_port_id(input_port_id, min_port_id=0, max_port_id=1)

        if len(args) == 0:
            kwargs["a" if input_port_id == 0 else "b"] = value
            return args, kwargs
        return replace_value_by_index(args, input_port_id, value), kwargs



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Tuple, Union

import numpy as np
import openvino.runtime as ov
from openvino import Type
from openvino.properties.hint import inference_precision

from nncf.common.engine import Engine
from nncf.openvino.graph.model_utils import model_has_state


class OVCompiledModelEngine(Engine):
    """
    Implementation of the engine to infer OpenVINO compiled model.

    OVCompiledModelEngine uses
    [OpenVINO Runtime](https://docs.openvino.ai/latest/openvino_docs_OV_UG_OV_Runtime_User_Guide.html)
    to infer the compiled model.
    """

    def __init__(self, compiled_model: ov.CompiledModel, stateful: bool):
        self.infer_request = compiled_model.create_infer_request()
        self.reset_state = stateful and hasattr(self.infer_request, "reset_state")

    def infer(
        self, input_data: Union[np.ndarray, List[np.ndarray], Tuple[np.ndarray], Dict[str, np.ndarray]]
    ) -> Dict[str, np.ndarray]:
        """
        Runs model on the provided input via OpenVINO Runtime.
        Returns the dictionary of model outputs by node names.

        :param input_data: Inputs for the model.
        :return output_data: Model's output.
        """
        if self.reset_state:
            self.infer_request.reset_state()

        model_outputs = self.infer_request.infer(input_data, share_inputs=True)

        output_data = {}
        for tensor, value in model_outputs.items():
            for tensor_name in tensor.get_names():
                output_data[tensor_name] = value
        return output_data


class OVNativeEngine(Engine):
    """
    Implementation of the engine for OpenVINO backend.

    OVNativeEngine uses
    [OpenVINO Runtime](https://docs.openvino.ai/latest/openvino_docs_OV_UG_OV_Runtime_User_Guide.html)
    to infer the model.
    """

    def __init__(self, model: ov.Model, use_fp32_precision: bool = True):
        """
        :param model: Model.
        :param use_fp32_precision: A flag that determines whether to force the engine to use FP32
            precision during inference.
        """
        config = None
        if use_fp32_precision:
            config = {inference_precision: Type.f32}
        ie = ov.Core()
        stateful = model_has_state(model)
        compiled_model = ie.compile_model(model, device_name="CPU", config=config)
        self.engine = OVCompiledModelEngine(compiled_model, stateful)

    def infer(
        self, input_data: Union[np.ndarray, List[np.ndarray], Tuple[np.ndarray], Dict[str, np.ndarray]]
    ) -> Dict[str, np.ndarray]:
        """
        Runs model on the provided input via OpenVINO Runtime.
        Returns the dictionary of model outputs by node names.

        :param input_data: Inputs for the model.
        :return output_data: Model's output.
        """
        return self.engine.infer(input_data)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch.nn

import nncf
from nncf.torch.dynamic_graph.scope import Scope


def get_module_by_scope(model: torch.nn.Module, scope: Scope) -> Optional[torch.nn.Module]:
    curr_module = model
    for scope_element in scope[1:]:  # omit first scope element which corresponds to base module
        if scope_element.calling_field_name is None:
            # The module used is being created in-place every time and never stored in the model,
            # happens for nn.Softmax in BERT implementations.
            return None

        next_module = curr_module._modules.get(scope_element.calling_field_name)
        if next_module is None:
            raise nncf.InternalError(
                "Could not find a {} module member in {} module of scope {} during node search".format(
                    scope_element.calling_field_name, scope_element.calling_module_class_name, str(scope)
                )
            )
        curr_module = next_module
    return curr_module



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from copy import deepcopy
from typing import Dict, List, Union

from nncf.common.graph import NNCFNodeName
from nncf.common.quantization.quantizer_setup import SingleConfigQuantizerSetup
from nncf.common.quantization.structs import QuantizerId
from nncf.common.quantization.structs import WeightQuantizerId
from nncf.torch.dynamic_graph.scope import Scope
from nncf.torch.graph.transformations.commands import ExtraCompressionModuleType
from nncf.torch.module_operations import UpdateWeight
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.quantization.layers import QUANTIZATION_MODULES
from nncf.torch.quantization.layers import BaseQuantizer
from nncf.torch.quantization.precision_constraints import HardwareQuantizationConstraints
from nncf.torch.quantization.structs import WeightQuantizerInfo
from nncf.torch.structures import NNCFExtraConfigStruct
from nncf.torch.utils import get_all_modules_by_type


class BasePrecisionInitParams:
    def __init__(self, user_init_args: NNCFExtraConfigStruct = None):
        self.user_init_args = user_init_args


class BasePrecisionInitializer:
    def __init__(
        self,
        algo: "ExperimentalQuantizationController",  # noqa: F821
        params: BasePrecisionInitParams,
        hw_precision_constraints: HardwareQuantizationConstraints = None,
    ):
        self._algo = algo
        self._model: NNCFNetwork = self._algo._model
        all_quantizers = algo.all_quantizations
        self._hw_precision_constraints = hw_precision_constraints
        self.original_precisions = {q_id: quantizer.num_bits for q_id, quantizer in all_quantizers.items()}
        self._quantizers_handler = WeightQuantizersHandler(
            self._model, self._algo.weight_quantizers, self._hw_precision_constraints
        )
        quantization_types = [class_type.__name__ for class_type in QUANTIZATION_MODULES.registry_dict.values()]
        self._weight_quantizations_by_execution_order = (
            self._quantizers_handler.get_weight_quantizers_in_execution_order_per_id()
        )

        self._all_quantizers_per_scope = get_all_modules_by_type(
            self._model.nncf.get_compression_modules_by_type(ExtraCompressionModuleType.EXTERNAL_QUANTIZER),
            quantization_types,
        )
        self._all_quantizers_per_scope.update(
            self._quantizers_handler.get_all_weight_quantizers_in_execution_order_per_scope()
        )

    def apply_init(self) -> SingleConfigQuantizerSetup:
        raise NotImplementedError

    @staticmethod
    def get_bitwidth_per_scope(quantizer_setup: SingleConfigQuantizerSetup) -> List[List[Union[int, str]]]:
        scope_vs_bitwidth = {}
        for qp in quantizer_setup.quantization_points.values():
            scope_vs_bitwidth[str(qp.insertion_point)] = qp.qconfig.num_bits
        sorted_scope_vs_bitwidth = OrderedDict(sorted(scope_vs_bitwidth.items(), key=lambda x: x[0]))
        full_bitwidth_per_scope = []
        for scope, bitwidth in sorted_scope_vs_bitwidth.items():
            full_bitwidth_per_scope.append([bitwidth, scope])
        return full_bitwidth_per_scope


class WeightQuantizersHandler:
    """
    Defines weight quantizers for precision initialization in the order of execution.
    """

    def is_wq_scope(self, scope: Scope) -> bool:
        return scope[-2].calling_module_class_name == UpdateWeight.__name__

    @staticmethod
    def get_owning_module_scope_from_wq_scope(wq_scope: Scope) -> Scope:
        retval = deepcopy(wq_scope)
        retval.pop()
        retval.pop()
        retval.pop()
        return retval

    def __init__(
        self,
        model: NNCFNetwork,
        weight_quantizers: Dict[WeightQuantizerId, WeightQuantizerInfo],
        constraints: HardwareQuantizationConstraints,
    ):
        self._wq_affected_module_node_name_vs_qid_dict = {k.target_node_name: k for k in weight_quantizers}
        self._quantizer_module_scope_vs_qid_dict: Dict[Scope, WeightQuantizerId] = {}
        self._skipped_quantized_weight_node_names = []
        self._skipped_weight_quantizers: Dict[WeightQuantizerId, BaseQuantizer] = {}
        self._weight_quantizers_in_execution_order_per_scope: Dict[Scope, BaseQuantizer] = OrderedDict()
        self._weight_quantizers_in_execution_order: Dict[WeightQuantizerId, BaseQuantizer] = OrderedDict()

        quantization_types = [class_type.__name__ for class_type in QUANTIZATION_MODULES.registry_dict.values()]
        weight_module_dict = model
        quantizers_in_execution_order_per_scope = get_all_modules_by_type(weight_module_dict, quantization_types)

        for scope, quantizer in quantizers_in_execution_order_per_scope.items():
            if self.is_wq_scope(scope):
                affected_module_scope = self.get_owning_module_scope_from_wq_scope(scope)
                affected_module_node = model.nncf.get_original_graph().get_op_nodes_in_scope(affected_module_scope)[0]
                if affected_module_node.node_name in self._wq_affected_module_node_name_vs_qid_dict:
                    qid = self._wq_affected_module_node_name_vs_qid_dict[affected_module_node.node_name]
                    if len(constraints.get_all_unique_bitwidths(qid)) != 1:
                        self._weight_quantizers_in_execution_order_per_scope[scope] = quantizer
                        self._weight_quantizers_in_execution_order[qid] = quantizer
                    else:
                        self._skipped_quantized_weight_node_names.append(affected_module_node.node_name)
                        self._skipped_weight_quantizers[qid] = quantizer

    def get_skipped_quantized_weight_node_names(self) -> List[NNCFNodeName]:
        return self._skipped_quantized_weight_node_names

    def get_all_weight_quantizers_in_execution_order_per_scope(self) -> Dict[Scope, BaseQuantizer]:
        return self._weight_quantizers_in_execution_order_per_scope

    def get_weight_quantizers_in_execution_order_per_id(self) -> Dict[WeightQuantizerId, BaseQuantizer]:
        return self._weight_quantizers_in_execution_order

    def get_quantizer_id_by_scope(self, scope: Scope) -> QuantizerId:
        affected_module_scope = self.get_owning_module_scope_from_wq_scope(scope)
        return self._wq_affected_module_node_name_vs_qid_dict[affected_module_scope]

    def get_skipped_weight_quantizers_per_id(self) -> Dict[QuantizerId, BaseQuantizer]:
        return self._skipped_weight_quantizers



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, List, Optional

from torch import nn

from nncf.torch.layers import NNCF_PADDING_VALUE_ATTR_NAME


class BaseOp(nn.Module):
    def __init__(self, op):
        super().__init__()
        self.op = op

    @property
    def operand(self):
        return self.op

    def forward(self, *inputs, **kwargs):
        return self.op(*inputs, **kwargs)


class UpdateInputs(BaseOp):
    """
    A module which updates inputs for a module
    fed to forward method call by operand call.
    """

    def __call__(self, _, inputs):
        return super().__call__(*inputs)


class UpdateParameter(BaseOp):
    """
    A module which updates the attribute by a given of a module
    fed to forward method call by operand call.
    """

    def __init__(self, param_name, op):
        super().__init__(op)
        self._param_name = param_name

    def __call__(self, module, _):
        if not hasattr(module, self._param_name):
            raise TypeError("{} should have {} attribute".format(type(module), self._param_name))
        value = getattr(module, self._param_name)
        result = super().__call__(value)
        setattr(module, self._param_name, result)


class UpdateWeight(UpdateParameter):
    """
    A module which updates `weight` attributes of a module
    fed to forward method call by operand call.
    """

    def __init__(self, op):
        super().__init__("weight", op)


class UpdateParameterList(BaseOp):
    """
    A module which updates attributes by a given list of names of a module fed to
    forward method call by operand call.
    """

    def __init__(self, param_names: List[str], op: Callable, is_optional_list: Optional[List[bool]] = None):
        super().__init__(op)
        self._param_names = param_names
        if is_optional_list is None:
            is_optional_list = [False for _ in param_names]
        self._is_optional_list = is_optional_list

    def __call__(self, module, _):
        param_values = []
        for param_name, is_optional in zip(self._param_names, self._is_optional_list):
            if not hasattr(module, param_name):
                if is_optional:
                    param_values.append(None)
                    continue
                raise TypeError("{} should have {} attribute".format(type(module), param_name))
            param_values.append(getattr(module, param_name))
        updated_kwargs = dict(zip(self._param_names, param_values))
        updated_values = super().__call__(**updated_kwargs)

        for param_name, updated_value in zip(self._param_names, updated_values):
            setattr(module, param_name, updated_value)


class UpdateWeightAndBias(UpdateParameterList):
    """
    A module which updates `weight` and `bias` attributes of a module
    fed to forward method call by operand call.
    """

    def __init__(self, op):
        super().__init__(["weight", "bias"], op)


class UpdateWeightAndOptionalBias(UpdateParameterList):
    """
    A module which updates `weight` and optionally `bias` attributes of a module
    fed to forward method call by operand call. If the module doesn't have bias attribute, None will be passed instead
    of it.
    """

    def __init__(self, op):
        super().__init__(["weight", "bias"], op, [False, True])


class UpdatePaddingValue(UpdateParameter):
    """
    A module which updates `nncf_padding` attributes of a module
    fed to forward method call by operand call. Eventually, that will be used to apply a custom padding value.
    """

    def __init__(self, op):
        super().__init__(NNCF_PADDING_VALUE_ATTR_NAME, op)


class UpdateNumGroups(UpdateParameter):
    """
    A module which updates `groups` attribute of a module
    fed to forward method call by operand call.
    """

    def __init__(self, op):
        super().__init__("groups", op)


class UpdatePadding(UpdateParameter):
    """
    A module which updates `padding` attribute of a module
    fed to forward method call by operand call.
    """

    def __init__(self, op):
        super().__init__("padding", op)


class UpdateBatchNormParams(UpdateParameterList):
    """
    A module which updates attribute of batch norm module
    fed to forward method call by operand call.
    """

    def __init__(self, op):
        super().__init__(["weight", "bias", "running_mean", "running_var"], op)


class UpdateLayerNormParams(UpdateParameterList):
    """
    A module which updates attribute of layer norm module
    fed to forward method call by operand call.
    """

    def __init__(self, op):
        super().__init__(["weight", "bias", "normalized_shape"], op)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional

import numpy as np
from safetensors.numpy import load_file as np_load_file
from safetensors.numpy import save_file as np_save_file

from nncf.tensor.definitions import TensorDeviceType
from nncf.tensor.functions import io as io
from nncf.tensor.functions.dispatcher import register_numpy_types
from nncf.tensor.functions.numpy_numeric import validate_device


def load_file(file_path: str, *, device: Optional[TensorDeviceType] = None) -> Dict[str, np.ndarray]:
    validate_device(device)
    return np_load_file(file_path)


@register_numpy_types(io.save_file)
def _(data: Dict[str, np.ndarray], file_path: str) -> None:
    return np_save_file(data, file_path)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Deque, List, Tuple, Union

import numpy as np
import tensorflow as tf

from nncf.common.tensor import NNCFTensor
from nncf.common.tensor import TensorElementsType
from nncf.common.tensor_statistics.collectors import MeanMinMaxStatisticCollector
from nncf.common.tensor_statistics.collectors import MeanPercentileStatisticCollector
from nncf.common.tensor_statistics.collectors import MedianMADStatisticCollector
from nncf.common.tensor_statistics.collectors import MinMaxStatisticCollector
from nncf.common.tensor_statistics.collectors import MixedMinMaxStatisticCollector
from nncf.common.tensor_statistics.collectors import PercentileStatisticCollector
from nncf.common.tensor_statistics.reduction import np_percentile_reduce_like
from nncf.tensorflow.tensor import TFNNCFTensor
from nncf.tensorflow.tensor_statistics.reduction import convert_rs_to_pt_type
from nncf.tensorflow.tensor_statistics.statistics import TFMedianMADTensorStatistic
from nncf.tensorflow.tensor_statistics.statistics import TFMinMaxTensorStatistic
from nncf.tensorflow.tensor_statistics.statistics import TFPercentileTensorStatistic


class TFNNCFCollectorTensorProcessor:
    """
    A realization of the processing methods for TFNNCFTensors.
    """

    @staticmethod
    def reduce_min(x: NNCFTensor, axis: Union[int, Tuple[int, ...], List[int]], keepdims: bool = False) -> NNCFTensor:
        return TFNNCFTensor(tf.reduce_min(x.tensor, axis=axis, keepdims=keepdims))

    @staticmethod
    def reduce_max(x: NNCFTensor, axis: Union[int, Tuple[int, ...], List[int]], keepdims: bool = False) -> NNCFTensor:
        return TFNNCFTensor(tf.reduce_max(x.tensor, axis=axis, keepdims=keepdims))

    @staticmethod
    def abs(x: NNCFTensor) -> NNCFTensor:
        return TFNNCFTensor(tf.math.abs(x.tensor))

    @staticmethod
    def min(x1: tf.Tensor, x2: tf.Tensor) -> NNCFTensor:
        return TFNNCFTensor(tf.math.minimum(x1.tensor, x2.tensor))

    @staticmethod
    def max(x1: tf.Tensor, x2: tf.Tensor) -> NNCFTensor:
        return TFNNCFTensor(tf.math.maximum(x1.tensor, x2.tensor))

    @staticmethod
    def mean(x: NNCFTensor, axis: Union[int, Tuple[int, ...], List[int]], keepdims=False) -> NNCFTensor:
        return TFNNCFTensor(tf.math.reduce_mean(x.tensor, axis=axis, keepdims=keepdims))

    @staticmethod
    def stack(x: Union[List[tf.Tensor], Deque[tf.Tensor]], axis: int = 0) -> NNCFTensor:
        x = [t.tensor for t in x]
        return TFNNCFTensor(tf.stack(x, axis=axis))

    @staticmethod
    def unstack(x: NNCFTensor, axis: int = 0) -> List[NNCFTensor]:
        tensor = x.tensor
        if list(tensor.shape) == []:
            tensor = tf.expand_dims(tensor, 0)
        tensor_list = tf.unstack(tensor, axis=axis)
        return [TFNNCFTensor(t) for t in tensor_list]

    @staticmethod
    def sum(tensor: NNCFTensor) -> TensorElementsType:
        return tf.reduce_sum(tensor.tensor).numpy()


class TFMinMaxStatisticCollector(MinMaxStatisticCollector):
    @staticmethod
    def _get_processor() -> TFNNCFCollectorTensorProcessor:
        return TFNNCFCollectorTensorProcessor()

    def _register_input(self, x: tf.Tensor):
        self._register_input_common(TFNNCFTensor(x))

    def _get_statistics(self) -> TFMinMaxTensorStatistic:
        return TFMinMaxTensorStatistic(self._min_values.tensor, self._max_values.tensor)


class TFMixedMinMaxStatisticCollector(MixedMinMaxStatisticCollector):
    @staticmethod
    def _get_processor() -> TFNNCFCollectorTensorProcessor:
        return TFNNCFCollectorTensorProcessor()

    def _register_input(self, x: tf.Tensor):
        self._register_input_common(TFNNCFTensor(x))

    def _get_statistics(self) -> TFMinMaxTensorStatistic:
        return TFMinMaxTensorStatistic(self._min_aggregate().tensor, self._max_aggregate().tensor)


class TFMeanMinMaxStatisticCollector(MeanMinMaxStatisticCollector):
    @staticmethod
    def _get_processor() -> TFNNCFCollectorTensorProcessor:
        return TFNNCFCollectorTensorProcessor()

    def _register_input(self, x: tf.Tensor):
        self._register_input_common(TFNNCFTensor(x))

    def _get_statistics(self) -> TFMinMaxTensorStatistic:
        return TFMinMaxTensorStatistic(self._min_aggregate().tensor, self._max_aggregate().tensor)


class TFMedianMADStatisticCollector(MedianMADStatisticCollector):
    def _register_input(self, x: tf.Tensor):
        self._samples.append(x.numpy())

    def _get_statistics(self) -> TFMedianMADTensorStatistic:
        self._reduction_shape = convert_rs_to_pt_type(self._samples[0].shape, self._reduction_shape)
        numpy_median, numpy_mad = self._prepare_statistics()
        median_tensor = tf.convert_to_tensor(np.array(numpy_median), dtype=tf.float32)
        mad_tensor = tf.convert_to_tensor(np.array(numpy_mad), dtype=tf.float32)
        return TFMedianMADTensorStatistic(median_tensor, mad_tensor)


class TFPercentileStatisticCollector(PercentileStatisticCollector):
    def _register_input(self, x: tf.Tensor):
        self._samples.append(x.numpy())

    def _get_statistics(self) -> TFPercentileTensorStatistic:
        self._reduction_shape = convert_rs_to_pt_type(self._samples[0].shape, self._reduction_shape)
        percentile_vs_values_dict = self._prepare_statistics()
        for key, val in percentile_vs_values_dict.items():
            percentile_vs_values_dict[key] = tf.convert_to_tensor(val, dtype=tf.float32)
        return TFPercentileTensorStatistic(percentile_vs_values_dict)


class TFMeanPercentileStatisticCollector(MeanPercentileStatisticCollector):
    def _register_input(self, x: tf.Tensor):
        x_np = x.numpy()
        for pct, values in self._all_pct_values.items():
            np_vals = np_percentile_reduce_like(x_np, convert_rs_to_pt_type(x_np.shape, self._reduction_shape), pct)
            tf_vals = tf.convert_to_tensor(np_vals, dtype=tf.float32)
            values.append(tf_vals)

    def _get_statistics(self) -> TFPercentileTensorStatistic:
        mean_percentile_values = {}
        for pct, values in self._all_pct_values.items():
            stacked_pct_vals = tf.stack(values)
            mean_percentile_values[pct] = tf.math.reduce_mean(stacked_pct_vals, axis=0)
        return TFPercentileTensorStatistic(mean_percentile_values)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.common.graph.transformations.commands import TargetType
from nncf.common.quantization.quantizer_setup import ActivationQuantizationInsertionPoint
from nncf.common.quantization.quantizer_setup import QuantizationInsertionPointBase
from nncf.common.quantization.quantizer_setup import WeightQuantizationInsertionPoint
from nncf.torch.graph.transformations.commands import PTTargetPoint


class PTTargetPointTranslator:
    @staticmethod
    def translate(qip: QuantizationInsertionPointBase) -> PTTargetPoint:
        if isinstance(qip, WeightQuantizationInsertionPoint):
            return PTTargetPoint(target_type=TargetType.OPERATION_WITH_WEIGHTS, target_node_name=qip.target_node_name)
        assert isinstance(qip, ActivationQuantizationInsertionPoint)
        input_port_id = qip.input_port_id
        if input_port_id is not None:
            return PTTargetPoint(
                target_type=TargetType.OPERATOR_PRE_HOOK,
                target_node_name=qip.target_node_name,
                input_port_id=input_port_id,
            )
        return PTTargetPoint(target_type=TargetType.OPERATOR_POST_HOOK, target_node_name=qip.target_node_name)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from typing import Optional

import torch
import torch.fx
from torch.ao.quantization.pt2e.duplicate_dq_pass import DuplicateDQPass
from torch.ao.quantization.pt2e.port_metadata_pass import PortNodeMetaForQDQ
from torch.ao.quantization.pt2e.qat_utils import _fold_conv_bn_qat
from torch.ao.quantization.pt2e.utils import _disallow_eval_train
from torch.fx import GraphModule
from torch.fx.passes.infra.pass_manager import PassManager

import nncf
from nncf.common.factory import NNCFGraphFactory
from nncf.common.logging import nncf_logger
from nncf.common.quantization.structs import QuantizationPreset
from nncf.data import Dataset
from nncf.experimental.torch.fx.quantization.backend_parameters import is_weight_compression_needed
from nncf.experimental.torch.fx.transformations import apply_quantization_transformations
from nncf.experimental.torch.fx.transformations import compress_post_quantize_transformation
from nncf.experimental.torch.fx.transformations import fq_weights_transformation
from nncf.parameters import BackupMode
from nncf.parameters import CompressWeightsMode
from nncf.parameters import ModelType
from nncf.parameters import QuantizationMode
from nncf.parameters import SensitivityMetric
from nncf.parameters import TargetDevice
from nncf.quantization.advanced_parameters import AdvancedCompressionParameters
from nncf.quantization.advanced_parameters import AdvancedQuantizationParameters
from nncf.quantization.algorithms.post_training.algorithm import PostTrainingQuantization
from nncf.quantization.algorithms.weight_compression.algorithm import WeightCompression
from nncf.scopes import IgnoredScope

DEFAULT_RANGE_TYPE = "mean_min_max"


def quantize_impl(
    model: torch.fx.GraphModule,
    calibration_dataset: Dataset,
    mode: Optional[QuantizationMode] = None,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_parameters: Optional[AdvancedQuantizationParameters] = None,
) -> torch.fx.GraphModule:
    """
    Implementation of the `quantize()` method for the Torch FX backend.
    """
    nncf_logger.warning(
        "Experimental Torch FX quantization backend is being used for the given torch.fx.GraphModule model."
        " Torch FX PTQ is an experimental feature, consider using Torch or OpenVino PTQ backends"
        " in case of errors or a poor model performance."
    )
    if target_device == TargetDevice.CPU_SPR:
        raise nncf.InternalError("target_device == CPU_SPR is not supported")
    if mode is not None:
        raise ValueError(f"mode={mode} is not supported")

    original_graph_meta = model.meta

    copied_model = deepcopy(model)

    quantization_algorithm = PostTrainingQuantization(
        preset=preset,
        target_device=target_device,
        subset_size=subset_size,
        fast_bias_correction=fast_bias_correction,
        model_type=model_type,
        ignored_scope=ignored_scope,
        advanced_parameters=advanced_parameters,
    )

    # To make it easier for bias correction algorithms.
    apply_quantization_transformations(copied_model)

    nncf_graph = NNCFGraphFactory.create(copied_model)
    quantized_model = quantization_algorithm.apply(copied_model, nncf_graph, dataset=calibration_dataset)

    if is_weight_compression_needed(advanced_parameters):
        compress_post_quantize_transformation(quantized_model)
    else:
        fq_weights_transformation(quantized_model)

    # Magic. Without this call compiled model
    # is not preformant
    quantized_model = GraphModule(quantized_model, quantized_model.graph)

    quantized_model = _fold_conv_bn_qat(quantized_model)
    pm = PassManager([DuplicateDQPass()])

    quantized_model = pm(quantized_model).graph_module
    pm = PassManager([PortNodeMetaForQDQ()])
    quantized_model = pm(quantized_model).graph_module

    quantized_model.meta.update(original_graph_meta)
    quantized_model = _disallow_eval_train(quantized_model)
    # Each transformation adds a duplicate tensor value to the model buffer.
    #  This step removes the duplicates tensor values from the buffer.
    quantized_model = GraphModule(quantized_model, quantized_model.graph)

    return quantized_model


def compress_weights_impl(
    model: torch.fx.GraphModule,
    dataset: Dataset,
    mode: CompressWeightsMode,
    ratio: float,
    group_size: int,
    ignored_scope: IgnoredScope,
    all_layers: bool,
    sensitivity_metric: SensitivityMetric,
    awq: bool,
    subset_size: int,
    scale_estimation: bool,
    gptq: bool,
    lora_correction: bool,
    backup_mode: BackupMode,
    advanced_parameters: Optional[AdvancedCompressionParameters] = None,
) -> torch.fx.GraphModule:
    """
    Implementation of the `compress_weights()` method for the Torch Fx backend.
    """

    compression_algorithm = WeightCompression(
        mode,
        ratio,
        group_size,
        ignored_scope,
        all_layers,
        sensitivity_metric,
        awq,
        subset_size,
        scale_estimation,
        gptq,
        lora_correction,
        backup_mode,
        advanced_parameters,
    )
    graph = NNCFGraphFactory.create(model)
    compressed_model = compression_algorithm.apply(model, graph, dataset=dataset)
    compressed_model = GraphModule(compressed_model, compressed_model.graph)
    compressed_model = _disallow_eval_train(compressed_model)

    return compressed_model



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains builder and controller class definitions for the knowledge distillation.
"""

from copy import deepcopy

from torch import nn

from nncf import NNCFConfig
from nncf.api.compression import CompressionLoss
from nncf.api.compression import CompressionScheduler
from nncf.api.compression import CompressionStage
from nncf.common.schedulers import BaseCompressionScheduler
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.api_marker import api
from nncf.config.schemata.defaults import KNOWLEDGE_DISTILLATION_SCALE
from nncf.config.schemata.defaults import KNOWLEDGE_DISTILLATION_TEMPERATURE
from nncf.torch.algo_selector import PT_COMPRESSION_ALGORITHMS
from nncf.torch.compression_method_api import PTCompressionAlgorithmBuilder
from nncf.torch.compression_method_api import PTCompressionAlgorithmController
from nncf.torch.graph.transformations.layout import PTTransformationLayout
from nncf.torch.knowledge_distillation.knowledge_distillation_loss import KnowledgeDistillationLoss
from nncf.torch.nncf_network import NNCFNetwork


@PT_COMPRESSION_ALGORITHMS.register("knowledge_distillation")
class KnowledgeDistillationBuilder(PTCompressionAlgorithmBuilder):
    def __init__(self, config: NNCFConfig, should_init: bool = True):
        super().__init__(config, should_init)
        self.kd_type = self._algo_config.get("type")
        self.scale = self._algo_config.get("scale", KNOWLEDGE_DISTILLATION_SCALE)
        self.temperature = self._algo_config.get("temperature", KNOWLEDGE_DISTILLATION_TEMPERATURE)
        if "temperature" in self._algo_config and self.kd_type == "mse":
            raise ValueError("Temperature shouldn't be stated for MSE Loss (softmax only feature)")

    def _get_transformation_layout(self, target_model: NNCFNetwork) -> PTTransformationLayout:
        self.original_model = deepcopy(target_model).nncf.get_clean_shallow_copy()
        for param in self.original_model.parameters():
            param.requires_grad = False
        return PTTransformationLayout()

    def _build_controller(self, model):
        return KnowledgeDistillationController(model, self.original_model, self.kd_type, self.scale, self.temperature)

    def initialize(self, model: NNCFNetwork) -> None:
        pass


@api()
class KnowledgeDistillationController(PTCompressionAlgorithmController):
    """
    Controller for the knowledge distillation in PT.
    """

    def __init__(
        self, target_model: NNCFNetwork, original_model: nn.Module, kd_type: str, scale: float, temperature: float
    ):
        super().__init__(target_model)
        original_model.train()
        self._scheduler = BaseCompressionScheduler()
        self._loss = KnowledgeDistillationLoss(
            target_model=target_model,
            original_model=original_model,
            kd_type=kd_type,
            scale=scale,
            temperature=temperature,
        )

    def compression_stage(self) -> CompressionStage:
        """
        Returns level of compression. Should be used on saving best checkpoints to distinguish between
        uncompressed, partially compressed and fully compressed models.
        """
        return CompressionStage.FULLY_COMPRESSED

    @property
    def scheduler(self) -> CompressionScheduler:
        return self._scheduler

    @property
    def loss(self) -> CompressionLoss:
        return self._loss

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        return NNCFStatistics()

    def distributed(self):
        pass



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.common.graph.patterns import GraphPattern
from nncf.tensorflow.graph.metatypes.tf_ops import TFAddOpMetatype
from nncf.tensorflow.graph.metatypes.tf_ops import TFBiasAddOpMetatype
from nncf.tensorflow.graph.metatypes.tf_ops import TFConv2DOpMetatype
from nncf.tensorflow.graph.metatypes.tf_ops import TFMatMulOpMetatype
from nncf.tensorflow.graph.metatypes.tf_ops import TFMulOpMetatype
from nncf.tensorflow.graph.metatypes.tf_ops import TFRelu6OpMetatype
from nncf.tensorflow.graph.metatypes.tf_ops import TFReluOpMetatype


def create_h_sigmoid_act() -> GraphPattern:
    main_pattern = GraphPattern()

    # ReLU version
    pattern = GraphPattern()

    input_pattern_node = pattern.add_node(label="*INPUT_NODE*", type=GraphPattern.NON_PATTERN_NODE_TYPE)
    add_node = pattern.add_node(label="ADD", type=TFAddOpMetatype.get_all_aliases())
    relu_node = pattern.add_node(label="RELU", type=TFReluOpMetatype.get_all_aliases())
    mul_node = pattern.add_node(label="TF_OP_MUL", type=TFMulOpMetatype.get_all_aliases())

    pattern.add_edge(input_pattern_node, add_node)
    pattern.add_edge(add_node, relu_node)
    pattern.add_edge(relu_node, mul_node)

    main_pattern.add_pattern_alternative(pattern)

    # ReLU6 version

    pattern = GraphPattern()

    input_pattern_node = pattern.add_node(label="*INPUT_NODE*", type=GraphPattern.NON_PATTERN_NODE_TYPE)
    add_node = pattern.add_node(label="ADD", type=TFAddOpMetatype.get_all_aliases())
    relu6_node = pattern.add_node(label="RELU6", type=TFRelu6OpMetatype.get_all_aliases())
    mul_node = pattern.add_node(label="TF_OP_MUL", type=TFMulOpMetatype.get_all_aliases())

    pattern.add_edge(input_pattern_node, add_node)
    pattern.add_edge(add_node, relu6_node)
    pattern.add_edge(relu6_node, mul_node)

    main_pattern.add_pattern_alternative(pattern)

    return main_pattern


def create_h_swish_act() -> GraphPattern:
    # TODO (vshampor): current approach with join_patterns is deficient since it does not allow to reliably
    #  connect nodes after a pattern has been joined. Along with the label and the type, the nodes created
    #  in the pattern must allow a "name" or "address" attribute, which must be a unique human readable
    #  string identifier of the node even if it has been joined multiple times, or perhaps each pattern
    #  after joining must return a list of output nodes so that these can be joined to later.
    #  Currently cannot specify h_swish in terms of h_sigmoid due to this.

    main_pattern = GraphPattern()

    # ReLU version
    pattern = GraphPattern()
    input_pattern_node = pattern.add_node(label="*INPUT_NODE*", type=GraphPattern.NON_PATTERN_NODE_TYPE)

    add_node = pattern.add_node(label="ADD", type=TFAddOpMetatype.get_all_aliases())
    relu_node = pattern.add_node(label="RELU", type=TFReluOpMetatype.get_all_aliases())
    mul_node = pattern.add_node(label="TF_OP_MUL", type=TFMulOpMetatype.get_all_aliases())

    pattern.add_edge(input_pattern_node, add_node)
    pattern.add_edge(add_node, relu_node)
    pattern.add_edge(relu_node, mul_node)

    mul_2_node = pattern.add_node(label="MULTIPLY", type=["Multiply", "Mul"])
    pattern.add_edge(input_pattern_node, mul_2_node)
    pattern.add_edge(mul_node, mul_2_node)
    main_pattern.add_pattern_alternative(pattern)

    # ReLU6 version
    pattern = GraphPattern()
    input_pattern_node = pattern.add_node(label="*INPUT_NODE*", type=GraphPattern.NON_PATTERN_NODE_TYPE)

    add_node = pattern.add_node(label="ADD", type=TFAddOpMetatype.get_all_aliases())
    relu6_node = pattern.add_node(label="RELU6", type=TFRelu6OpMetatype.get_all_aliases())
    mul_node = pattern.add_node(label="TF_OP_MUL", type=TFMulOpMetatype.get_all_aliases())

    pattern.add_edge(input_pattern_node, add_node)
    pattern.add_edge(add_node, relu6_node)
    pattern.add_edge(relu6_node, mul_node)

    mul_2_node = pattern.add_node(label="MULTIPLY", type="Multiply")
    pattern.add_edge(input_pattern_node, mul_2_node)
    pattern.add_edge(mul_node, mul_2_node)
    main_pattern.add_pattern_alternative(pattern)

    return main_pattern


def create_matmul_biasadd_pattern() -> GraphPattern:
    pattern = GraphPattern()

    matmul_node = pattern.add_node(label="MATMUL", type=TFMatMulOpMetatype.get_all_aliases())
    biasadd_node = pattern.add_node(label="BIASADD", type=TFBiasAddOpMetatype.get_all_aliases())
    pattern.add_edge(matmul_node, biasadd_node)

    return pattern


def create_conv2d_biasadd_pattern() -> GraphPattern:
    pattern = GraphPattern()

    conv2d_node = pattern.add_node(label="CONV2D", type=TFConv2DOpMetatype.get_all_aliases())
    biasadd_node = pattern.add_node(label="BIASADD", type=TFBiasAddOpMetatype.get_all_aliases())
    pattern.add_edge(conv2d_node, biasadd_node)

    return pattern



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from copy import deepcopy
from typing import Callable, List, Tuple

import torch
from torch.nn import DataParallel

from nncf.common.graph.definitions import MODEL_CONST_OP_NAME
from nncf.common.graph.layer_attributes import BaseLayerAttributes
from nncf.common.graph.layer_attributes import WeightedLayerAttributes
from nncf.common.logging import nncf_logger
from nncf.common.utils.debug import is_debug
from nncf.torch.dynamic_graph.context import TracingContext
from nncf.torch.dynamic_graph.context import get_current_context
from nncf.torch.dynamic_graph.layer_attributes_handlers import get_layer_attributes_from_args_and_kwargs
from nncf.torch.dynamic_graph.op_input_processing import OperatorInput
from nncf.torch.dynamic_graph.operation_address import OperationAddress
from nncf.torch.dynamic_graph.patch_pytorch_state import PATCHING_STATE
from nncf.torch.dynamic_graph.structs import NamespaceTarget
from nncf.torch.dynamic_graph.structs import PatchedOperatorInfo
from nncf.torch.dynamic_graph.trace_functions import forward_trace_only
from nncf.torch.dynamic_graph.trace_functions import make_tensor_metas
from nncf.torch.dynamic_graph.trace_functions import trace_tensors
from nncf.torch.dynamic_graph.trace_tensor import TracedParameter
from nncf.torch.layer_utils import _NNCFModuleMixin
from nncf.torch.layers import ITERATION_MODULES

_IGNORED_SCOPES = []


def _warn_data_parallel():
    if getattr(_warn_data_parallel, "warned_once", False):
        return
    _warn_data_parallel.warned_once = True
    nncf_logger.warning(
        "You are using DataParallel, which may cause significant performance issues with dynamic graph "
        "building. Consider using distributed training (DistributedDataParallel) instead."
    )


def ignore_scope(cls):
    if cls not in _IGNORED_SCOPES:
        _IGNORED_SCOPES.append(cls)
    return cls


def wrap_operator(operator, operator_info: PatchedOperatorInfo):
    """
    Wraps the input callable object (`operator`) with the functionality that allows the calls to this object
    to be tracked by the currently set global TracingContext. The wrapped functions can be then intercepted,
    their arguments and return values modified arbitrarily and, for functions that correspond to operations on
    tensors in a DNN, their general position and address in the DNN's model control flow graph can be established.

    :param: operator: A callable object to be wrapped.
    :param: operator_info (PatchedOperatorInfo): An informational struct containing the specifics of wrapping
            the `operator` in question.

    :return: The wrapped version of `operator` that, without a TracingContext, performs functionally the same as
             the unwrapped version, but within a TracingContext is able to be tracked and hooked.
    """
    # do not wrap function twice
    _orig_op = getattr(operator, "_original_op", None)
    if _orig_op is not None:
        nncf_logger.debug(f"Operator: {_orig_op.__name__} is already wrapped")
        return operator

    @functools.wraps(operator)
    def wrapped(*args, **kwargs):
        if not PATCHING_STATE.operators_are_wrapped:
            # If operators are not supposed to be wrapped, skip the wrapper logic
            return operator(*args, **kwargs)

        ctx = get_current_context()
        if not ctx or getattr(ctx, "in_operator", False) or not ctx.is_tracing:
            op1 = operator(*args, **kwargs)
            return op1

        ctx.in_operator = True

        try:
            if operator_info.skip_trace:
                result = operator(*args, **kwargs)
            elif ctx.is_forwarding:
                result = forward_trace_only(operator, *args, **kwargs)
            else:
                op_name = operator_info.name
                op_address = ctx.get_caller_context(op_name)
                ctx.register_operator_call(op_address.operator_name, op_address.scope_in_model)

                if ctx.elastic_depth and ctx.in_skipped_block:
                    result = ctx.tensor_cache
                else:
                    result = _execute_op(op_address, operator_info, operator, ctx, *args, **kwargs)

                str_op_address = str(op_address)
                if str_op_address in ctx.end_node_name_of_skipped_block:
                    assert ctx.in_skipped_block is True
                    ctx.in_skipped_block = False
                if str_op_address in ctx.start_node_name_of_skipped_block:
                    assert ctx.in_skipped_block is False, "skipping of overlapping blocks"
                    ctx.in_skipped_block = True
                    ctx.tensor_cache = result
        except:
            # Looks like the __repr__ call made during IDE debug to display tensor contents does not exit properly,
            # but instead throws an exception. This try...except block handles such a situation.
            # Otherwise the context is stuck in the "in_operator == True" state.
            ctx.in_operator = False
            raise

        ctx.in_operator = False
        return result

    wrapped._original_op = operator
    wrapped._operator_namespace = operator_info.operator_namespace
    return wrapped


def wrap_module_call(module_call):
    from nncf.torch.dynamic_graph.patch_pytorch import ORIGINAL_OPERATORS

    NAMES_ORIGINAL_OPERATORS = [op.name for op in ORIGINAL_OPERATORS]

    @functools.wraps(module_call)
    def wrapped(self, *args, **kwargs):
        from nncf.torch.dynamic_graph.patch_pytorch import unpatching_module_call

        # If called on a model compiled by torch dynamo, we unpatch torch operators and invoke original module call
        if "_torchdynamo_orig_callable" in self.forward.__dict__:
            return unpatching_module_call(self, *args, **kwargs)

        ctx = get_current_context()
        if not ctx or self.__class__ in _IGNORED_SCOPES:
            if isinstance(self, DataParallel):
                _warn_data_parallel()
            return module_call(self, *args, **kwargs)
        ctx.push_scope(self)
        is_nncf_layer = isinstance(self, _NNCFModuleMixin)
        if is_nncf_layer:
            op_name = self.op_func_name
        else:
            op_name = self.__class__.__name__
        is_layer_or_op = op_name in NAMES_ORIGINAL_OPERATORS or is_nncf_layer
        op_address = ctx.get_caller_context(op_name)
        str_op_address = str(op_address)
        if ctx.elastic_depth and is_layer_or_op and ctx.in_skipped_block:
            ctx.register_operator_call(op_address.operator_name, op_address.scope_in_model)
            retval = ctx.tensor_cache
            if str_op_address in ctx.end_node_name_of_skipped_block:
                assert ctx.in_skipped_block is True
                ctx.in_skipped_block = False
            if str_op_address in ctx.start_node_name_of_skipped_block:
                assert ctx.in_skipped_block is False, "skipping of overlapping blocks"
                ctx.in_skipped_block = True
        else:
            retval = module_call(self, *args, **kwargs)

        if type(self).__name__ in ITERATION_MODULES.registry_dict:
            ctx.reset_operator_call_count_in_scope(ctx.scope)
        ctx.pop_scope()
        return retval

    return wrapped


def _execute_op(
    op_address: OperationAddress,
    operator_info: PatchedOperatorInfo,
    operator: Callable,
    ctx: TracingContext,
    *args,
    **kwargs,
):
    op_name = operator_info.name

    op_input = OperatorInput(list(args), kwargs)
    op_input = _process_parameters(op_input, ctx)
    processed_input = ctx.execute_pre_hooks(op_address, op_input)
    args = tuple(processed_input.op_args)
    kwargs = processed_input.op_kwargs
    result = operator(*args, **kwargs)
    node = None
    if isinstance(result, type(NotImplemented)):
        nncf_logger.debug("Operation {} returned NotImplemented".format(op_name))
    elif ctx.trace_dynamic_graph:
        tensor_metas = make_tensor_metas(processed_input)
        node = ctx.find_operator_node(tensor_metas, op_address)
        if node is None:
            layer_attrs, ignored_algos = _collect_module_attrs_and_ignored_algorithms(ctx, op_name, args, kwargs)
            is_called_inside_nncf_module = isinstance(ctx.get_current_module(), _NNCFModuleMixin)
            node = ctx.maybe_add_node(
                processed_input, tensor_metas, op_address, layer_attrs, ignored_algos, is_called_inside_nncf_module
            )
        if is_debug() and node is not None:
            ctx.register_node_call(node)

    result = trace_tensors(result, node, ctx)
    result = ctx.execute_post_hooks(op_address, result)
    return result


def _collect_module_attrs_and_ignored_algorithms(
    ctx: TracingContext, op_name: str, args, kwargs
) -> Tuple[BaseLayerAttributes, List[str]]:
    ignored_algos = []
    layer_attrs = get_layer_attributes_from_args_and_kwargs(op_name, args, kwargs)

    curr_module = ctx.get_current_module()
    if curr_module is not None:
        if isinstance(curr_module, _NNCFModuleMixin):
            ignored_algos = deepcopy(curr_module.ignored_algorithms)

        if (
            isinstance(layer_attrs, WeightedLayerAttributes)
            and hasattr(curr_module, "weight_g")
            and hasattr(curr_module, "weight_v")
        ):
            # torch.nn.utils.weight_norm replaces weight with weight_g and weight_v
            layer_attrs.weight_requires_grad = curr_module.weight_g.requires_grad

    return layer_attrs, ignored_algos


@functools.partial(wrap_operator, operator_info=PatchedOperatorInfo(MODEL_CONST_OP_NAME, NamespaceTarget.EXTERNAL))
def process_parameter_fn(x: torch.nn.Parameter) -> torch.nn.Parameter:
    """
    The identity binding function to trace and apply hooks to parameters.

    :param x: A parameter.
    :return: A parameter.
    """
    return x


def _process_parameters(operator_inputs: OperatorInput, ctx: TracingContext) -> OperatorInput:
    """
    Process model parameters into operator inputs applying registered hooks to them. The function guarantees
    that the parameter is processed once.

    :param operator_inputs: The operator inputs.
    :param ctx: The compression context.
    :return: The operator inputs with processed parameters.
    """
    if ctx.in_parameter_trace:
        return operator_inputs

    in_op = getattr(ctx, "in_operator", False)
    ctx.in_operator = False

    for idx in range(len(operator_inputs)):
        traced_parameter = operator_inputs[idx]
        if not isinstance(traced_parameter, TracedParameter):
            continue

        processed_parameter = ctx.get_processed_parameter(traced_parameter.name)
        if processed_parameter is not None:
            operator_inputs[idx] = processed_parameter
            continue

        if traced_parameter.tensor_meta is not None:
            continue

        in_parameter_trace = getattr(ctx, "in_parameter_trace", False)
        ctx.in_parameter_trace = True
        is_reused = traced_parameter.is_reused
        processed_parameter = process_parameter_fn(traced_parameter)
        operator_inputs[idx] = processed_parameter
        if is_reused:
            ctx.register_processed_parameter(traced_parameter.name, processed_parameter)
        ctx.in_parameter_trace = in_parameter_trace

    ctx.in_operator = in_op
    return operator_inputs


def wrap_parameters(model: torch.nn.Module):
    """
    Wrap model parameters inplace by adding tracing capabilities.

    :param model: A model.
    """
    ctx = get_current_context()
    for name, param in model.named_parameters():
        if name.startswith("_nncf"):
            # Exclude parameters in modules which added by NNCF.
            continue
        is_reused = name in ctx.reused_parameters
        tt = TracedParameter.from_torch_parameter(param, name, is_reused)
        ctx.register_traced_tensor(tt)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from nncf.experimental.common.tensor_statistics.collectors import BatchMeanReducer
from nncf.experimental.common.tensor_statistics.collectors import MeanAggregator
from nncf.experimental.common.tensor_statistics.collectors import MeanPerChReducer
from nncf.experimental.common.tensor_statistics.collectors import NoopAggregator
from nncf.experimental.common.tensor_statistics.collectors import RawReducer
from nncf.experimental.common.tensor_statistics.collectors import ShapeAggregator
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.common.tensor_statistics.statistics import MeanTensorStatistic
from nncf.experimental.common.tensor_statistics.statistics import RawTensorStatistic


def get_mean_statistic_collector(
    num_samples: int, channel_axis: int, window_size: Optional[int] = None, inplace: bool = True
) -> TensorCollector:
    """
    Mean statistic collector builder.

    :param num_samples: Maximum number of samples to collect.
    :param channel_axis: Channel axis to use during reduction phase.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :param inplace: Whether the mean reducer should be calculated inplace or out of place.
    :return: Mean statistic collector.
    """
    inplace = False
    if channel_axis == 0:
        reducer = BatchMeanReducer(inplace)
    else:
        reducer = MeanPerChReducer(channel_axis=channel_axis, inplace=inplace)
    raw_reducer = RawReducer()

    kwargs = {
        "num_samples": num_samples,
        "window_size": window_size,
    }

    aggregate_mean = MeanAggregator(**kwargs)
    aggregate_shape = ShapeAggregator()

    collector = TensorCollector(MeanTensorStatistic)
    collector.register_statistic_branch(MeanTensorStatistic.MEAN_STAT, reducer, aggregate_mean)
    collector.register_statistic_branch(MeanTensorStatistic.SHAPE_STAT, raw_reducer, aggregate_shape)
    return collector


def get_raw_stat_collector(num_samples: int) -> TensorCollector:
    """
    Raw statistic collector builder.

    :param num_samples: Maximum number of samples to collect.
    :return: Raw statistic collector.
    """
    reducer = RawReducer()
    aggregator = NoopAggregator(num_samples)

    collector = TensorCollector(RawTensorStatistic)
    collector.register_statistic_branch(RawTensorStatistic.VALUES_STATS, reducer, aggregator)
    return collector



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Dict, List, Optional, Type, TypeVar

import nncf
from nncf.common import factory
from nncf.common.factory import NNCFGraphFactory
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.logging.track_progress import track
from nncf.common.scopes import should_consider_scope
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.data import Dataset
from nncf.experimental.torch.sparsify_activations.target_scope import TargetScope
from nncf.experimental.torch.sparsify_activations.target_scope import get_target_node_names_from_target_scope
from nncf.scopes import IgnoredScope
from nncf.scopes import get_ignored_node_names_from_ignored_scope
from nncf.torch.model_creation import is_wrapped_model
from nncf.torch.model_creation import wrap_model

TModel = TypeVar("TModel")


class SparsifyActivationsAlgoBackend(ABC):
    """
    Abstract class for activation sparsification algorithm backend.
    """

    CALIBRATION_TRACKING_DESC = "Conducting Activations Sparsifier Calibration"

    @staticmethod
    def do_inference(model: TModel, dataset: Dataset):
        """
        Conducts model inference on given dataset to calibrate the activation sparsifiers.

        :param model: The model with activation sparsifiers.
        :param dataset: The calibration dataset to update the sparsifiers.
        """
        engine = factory.EngineFactory.create(model)
        for input_data in track(
            dataset.get_inference_data(),
            total=dataset.get_length(),
            description=SparsifyActivationsAlgoBackend.CALIBRATION_TRACKING_DESC,
        ):
            engine.infer(input_data)

    @property
    @abstractmethod
    def supported_metatypes(self) -> List[Type[OperatorMetatype]]:
        """
        Property for the backend-specific metatypes for supported layers.
        """

    @abstractmethod
    def insert_sparsifiers(
        self,
        model: TModel,
        graph: NNCFGraph,
        target_sparsity_by_node: Dict[NNCFNode, float],
    ) -> TModel:
        """
        Inserts the activation sparsifiers to the model.

        :param model: The model to conduct activation sparsification.
        :param graph: The model's NNCF graph.
        :param target_sparsity_by_node: The target sparsity level for the input activation in each given node layer.
        :return: The model with inserted activation sparsifiers.
        """

    @abstractmethod
    def calibrate_sparsifiers(self, model: TModel, graph: NNCFGraph, dataset: Dataset) -> TModel:
        """
        Calibrates the thresholds in the activation sparsifiers.

        :param model: The model with inserted activation sparsifiers.
        :param graph: The model's NNCF graph.
        :param dataset: The calibration dataset to update the thresholds in the sparsifiers.
        :return: The model with calibrated activation sparsifiers.
        """


class SparsifyActivationsAlgorithm:
    """
    Implementation of activation sparsification algorithm.
    """

    def __init__(
        self,
        target_sparsity_by_scope: Dict[TargetScope, float],
        ignored_scope: IgnoredScope,
    ):
        """
        :param target_sparsity_by_scope: A dictionary that defines the target sparsity level for specified layers.
        :param ignored_scope: An ignored scope that defines the list of model control flow
            graph nodes to be ignored during activation sparsification.
        """
        self._target_sparsity_by_scope = target_sparsity_by_scope
        self._ignored_scope = ignored_scope
        self._backend_entity: SparsifyActivationsAlgoBackend = None

    @property
    def available_backends(self) -> List[BackendType]:
        """
        Supported backends for this algorithm.
        """
        return [BackendType.TORCH]

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        dataset: Dataset,
    ) -> TModel:
        """
        Applies the algorithm to the given model.

        :param model: The model to be sparsified.
        :param graph: The model's NNCF graph.
        :param dataset: The dataset to calibrate the activation sparsifiers.
        :return: The sparsified model.
        """
        self._set_backend_entity(model)
        target_sparsity_by_node = self._get_target_sparsity_by_node(graph)
        sparse_model = self.do_sparsification(model, graph, target_sparsity_by_node, dataset)
        return sparse_model

    def do_sparsification(
        self,
        model: TModel,
        graph: NNCFGraph,
        target_sparsity_by_node: Dict[NNCFNode, float],
        dataset: Dataset,
    ):
        """
        Transforms the model into a sparsified one with node-specific target activation sparsity levels.

        :param model: The model to be sparsified.
        :param graph: The model's NNCF graph.
        :param target_sparsity_by_node: A dictionary that defines the target sparsity level
            for specified node layers.
        :param dataset: The dataset to calibrate the activation sparsifiers.
        :return: The sparsified model.
        """
        model = self._backend_entity.insert_sparsifiers(model, graph, target_sparsity_by_node)
        model = self._backend_entity.calibrate_sparsifiers(model, graph, dataset)
        return model

    def _set_backend_entity(self, model: TModel) -> None:
        """
        Creates a helper class with a backend-specific logic of the algorithm.

        :param model: Backend-specific input model.
        """
        model_backend = get_backend(model)
        if model_backend == BackendType.TORCH:
            from nncf.experimental.torch.sparsify_activations.torch_backend import PTSparsifyActivationsAlgoBackend

            self._backend_entity = PTSparsifyActivationsAlgoBackend()
        else:
            raise nncf.UnsupportedBackendError(
                f"{model_backend.value} backend is not supported for `sparsify_activations`."
            )

    def _get_target_sparsity_by_node(self, graph: NNCFGraph) -> Dict[NNCFNode, float]:
        """
        Collects nodes in the model's graph corresponding to the layers for sparsification.

        :param graph: NNCFGraph instance.
        :return: A dictionary with nodes and the corresponding target sparsity level.
        """
        supported_metatypes = self._backend_entity.supported_metatypes
        ignored_names = get_ignored_node_names_from_ignored_scope(
            self._ignored_scope, graph, strict=self._ignored_scope.validate
        )
        target_sparsity_by_node = {}
        for scope, target_sparsity in self._target_sparsity_by_scope.items():
            target_names = get_target_node_names_from_target_scope(scope, graph, strict=scope.validate)
            for node_name in target_names:
                node = graph.get_node_by_name(node_name)
                if node.metatype not in supported_metatypes or not should_consider_scope(
                    node.node_name, ignored_scopes=ignored_names
                ):
                    continue
                if node in target_sparsity_by_node:
                    raise nncf.ValidationError(
                        f'"{node.node_name}" is matched by multiple items in `target_sparsity_by_scope`.'
                    )
                target_sparsity_by_node[node] = target_sparsity
        if not target_sparsity_by_node:
            raise nncf.ValidationError("No layers to conduct activation sparsification.")
        return target_sparsity_by_node


def sparsify_activations(
    model: TModel,
    dataset: Dataset,
    target_sparsity_by_scope: Dict[TargetScope, float],
    ignored_scope: Optional[IgnoredScope] = None,
) -> TModel:
    """
    Post-training activation sparsification on the given model.

    This algorithm sparsifies the input activations in supported layers based on a calibration
    dataset. The goal is to zero out neurons with small activation values around 0, thereby
    roughly achieving the target sparsity at a statistical level.

    Note that currently only linear layers are supported.

    :param model: The model to be sparsified.
    :param dataset: The dataset to calibrate the activation sparsifiers.
    :param target_sparsity_by_scope: Defines the target activation sparsity level
        for specified layers. For each item, the key is an instance of `TargetScope` class
        representing the layers to match in the model's NNCF graph; the corresponding value
        is a float number in the range [0, 1] representing the target sparsity level.

        Example:
        ..  code-block:: python
            {
                # Target sparsity is 60% for node "Dummy/Linear[layer]/linear_0" in the model graph
                TargetScope(names=["Dummy/Linear[layer]/linear_0"]): 0.6,
                # Target sparsity is 30% for the layers whose name contains "up_proj" or "down_proj".
                TargetScope(patterns=[".*up_proj.*", ".*down_proj.*"]): 0.3,
            }

    :param ignored_scope: Optional. It defines the nodes in the model graph that should be
        ignored during activation sparsification. Note that unsupported layer types are already
        filtered out internally, so there is no need to mention them in `ignored_scope`.
    :return: The sparsified model.
    """

    for scope, target_sparsity in target_sparsity_by_scope.items():
        if target_sparsity < 0.0 or target_sparsity > 1.0:
            raise ValueError(f'Target sparsity for scope "{scope}" should be in range [0, 1].')

    if ignored_scope is None:
        ignored_scope = IgnoredScope()

    backend = get_backend(model)
    if backend == BackendType.TORCH and not is_wrapped_model(model):
        example_input = next(iter(dataset.get_inference_data()))
        model = wrap_model(
            model,
            example_input=example_input,
            trace_parameters=True,
        )

    algorithm = SparsifyActivationsAlgorithm(target_sparsity_by_scope, ignored_scope)

    graph = NNCFGraphFactory.create(model)
    sparse_model = algorithm.apply(model, graph, dataset)
    return sparse_model



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.torch.dynamic_graph.scope import Scope


class OperationAddress:
    def __init__(self, operator_name: str, scope_in_model: Scope, call_order: int):
        self.operator_name = operator_name
        self.scope_in_model = scope_in_model
        self.call_order = call_order

    def __eq__(self, other: "OperationAddress"):
        return (
            isinstance(other, OperationAddress)
            and (self.operator_name == other.operator_name)
            and (self.scope_in_model == other.scope_in_model)
            and (self.call_order == other.call_order)
        )

    def __str__(self):
        return str(self.scope_in_model) + "/" + self.operator_name + "_" + str(self.call_order)

    def __hash__(self):
        return hash((self.operator_name, self.scope_in_model, self.call_order))

    @staticmethod
    def from_str(s: str):
        scope_and_op, _, call_order_str = s.rpartition("_")
        scope_str, _, op_name = scope_and_op.rpartition("/")
        return OperationAddress(op_name, Scope.from_str(scope_str), int(call_order_str))



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from typing import Any, Dict, List, Optional, Tuple

import tensorflow as tf
from tensorflow.lite.python.util import get_grappler_config as _get_grappler_config
from tensorflow.lite.python.util import run_graph_optimizations as _run_graph_optimizations
from tensorflow.python.framework import convert_to_constants as _convert_to_constants
from tensorflow.python.keras.saving import saving_utils as _saving_utils

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph.layer_attributes import Dtype
from nncf.experimental.tensorflow.graph.node_attributes import TFNodeAttributes
from nncf.experimental.tensorflow.graph.node_attributes import TFWeightedNodeAttributes
from nncf.tensorflow.graph.converter import TFModelConverter
from nncf.tensorflow.graph.metatypes.common import ALL_LAYER_METATYPES_WITH_WEIGHTS
from nncf.tensorflow.graph.metatypes.matcher import get_op_metatype


class TensorFlowGraphBuilder:
    """
    Converts given Keras model to the tf.Graph.
    """

    def __init__(self, model: tf.keras.Model, input_signature):
        """
        Initializes the `TensorFlowGraphBuilder`.

        :param model: The instance of the `tf.keras.Model` class.
        :param input_signature: A structure of the `tf.TensorSpec` objects specifying the
            inputs to the model.
        """
        self._model = model
        self._input_signature = tf.nest.flatten(input_signature)

    def build(self, graph_optimizers: List[str]):
        """
        :param graph_optimizers: A list of the strings that represent
            the list of optimizers.
                - `constfold`
                - `layout`
                - `pruning`
                - `arithmetic`
                - `dependency`
                - `function`
            Look at the [guide](https://www.tensorflow.org/guide/graph_optimization) for more information.
        :return: The optimized tf.Graph.
        """
        # Step 1: Freeze Keras model to frozen graph

        func = _saving_utils.trace_model_call(self._model, self._input_signature)
        concrete_func = func.get_concrete_function()
        frozen_func, graph_def = _convert_to_constants.convert_variables_to_constants_v2_as_graph(
            concrete_func, lower_control_flow=False
        )
        # List of input tensors
        input_tensors = [tensor for tensor in frozen_func.inputs if tensor.dtype != tf.dtypes.resource]
        # List of output tensors
        output_tensors = frozen_func.outputs

        # Step 2: Run a Grappler pass to oprimize the TensorFlow graph.

        # Creates a ConfigProto for configuring Grappler
        grappler_config = _get_grappler_config(graph_optimizers)
        # Skip running grappler when there are no optimizers to run. If not,
        # grappler will run with the default optimizer set and it will lead to
        # causing an unexpected behavior.
        if grappler_config.graph_options.rewrite_options.optimizers:
            graph_def = _run_graph_optimizations(
                graph_def, input_tensors, output_tensors, config=grappler_config, graph=frozen_func.graph
            )

        # Step 3: Convert the GraphDef to a tf.Graph
        with tf.Graph().as_default() as graph:
            tf.graph_util.import_graph_def(graph_def, name="")

        return graph, input_tensors, output_tensors


class NodeDesc:
    """
    Contains description of a node in the TensorFlow `FuncGraph`.
    This information is required for `NNCFNode` creation.
    """

    def __init__(
        self,
        op_name: str,
        op_type_name: str,
        metatype,
        is_shared: bool,
        resource_name: Optional[str] = None,
        attrs: Optional[Any] = None,
    ):
        """
        Initializes description of a node.

        :param op_name: Name of a node in the `FuncGraph`.
        :param op_type_name: Type of operation.
        :param metatype: NNCF meta type which corresponds to operation.
        :param is_shared: `True` if the node is shared, `False` otherwise.
        :param resource_name: Name of a node that contains weights of operation.
        :param attrs: Attributes of a node.
        """
        self.op_name = op_name
        self.op_type_name = op_type_name
        self.metatype = metatype
        self.is_shared = is_shared
        self.resource_name = resource_name
        self.attrs = attrs


class EdgeDesc:
    """
    Contains description of an edge in the TensorFlow `FuncGraph`.
    This information is required for `NNCFGraphEdge` creation.
    """

    def __init__(
        self,
        producer_op_name: str,
        output_port_id: int,
        consumer_op_name: str,
        input_port_id: int,
        tensor_shape: List[int],
        tensor_dtype: Dtype,
    ):
        """
        Initializes description of an edge.

        :param producer_op_name: Name of the node where the edge comes out.
        :param output_port_id: Output port id.
        :param consumer_op_name: Name of the node where the edge comes in.
        :param input_port_id: Input port id.
        :param tensor_shape: Shape of the tensor corresponding to the edge.
        :param tensor_dtype: Dtype of tensor.
        """
        self.producer_op_name = producer_op_name
        self.output_port_id = output_port_id
        self.consumer_op_name = consumer_op_name
        self.input_port_id = input_port_id
        self.tensor_shape = tensor_shape
        self.tensor_dtype = tensor_dtype


class SubclassedConverter(TFModelConverter):
    """
    Converts the Keras model, which was created via `tf.keras.Model`
    subclass, to the `NNCFGraph`.

    Uses the frozen TF graph for the provided model with applied
    graph optimizers to create `NNCFGraph`
    """

    def __init__(
        self,
        model: tf.keras.Model,
        input_signature: List[tf.TensorSpec],
        training: bool = False,
        graph_optimizers: Optional[List[str]] = None,
    ):
        """
        Initializes the subclassed converter.

        :param model: Instance of the `tf.keras.Model` class.
        :param input_signature: A list of `tf.TensorSpec` objects specifying the
            inputs to the model.
        :param training: Mode of the model.
        :param graph_optimizers: A list of the strings that represent
            the list of optimizers.
                - `constfold`
                - `layout`
                - `pruning`
                - `arithmetic`
                - `dependency`
                - `function`
            Look at the [guide](https://www.tensorflow.org/guide/graph_optimization) for more information.
            If `graph_optimizers` is `None` `dependency` optimizer will be used.
        """
        self._model = model
        self._input_signature = input_signature
        self._training = training

        if graph_optimizers is None:
            graph_optimizers = ["dependency"]
        self._graph_optimizers = graph_optimizers

        self._tfgraph_builder = TensorFlowGraphBuilder(self._model, self._input_signature)

    def convert(self) -> NNCFGraph:
        """
        Converts the Keras model to the `NNCFGraph` object.

        :return: The `NNCFGraph` object that represents the Keras model
            for compression algorithms.
        """
        tfgraph, input_tensors, _ = self._tfgraph_builder.build(self._graph_optimizers)

        input_op_names = [tensor.op.name for tensor in input_tensors]
        node_descs, edge_descs = SubclassedConverter._collect_tfgraph_descs(tfgraph, input_op_names)

        nncf_graph = SubclassedConverter._create_nncf_graph_from_descs(node_descs, edge_descs)

        return nncf_graph

    @staticmethod
    def _create_nncf_graph_from_descs(node_descs: List[NodeDesc], edge_descs: List[EdgeDesc]) -> NNCFGraph:
        """
        Creates the NNCF graph from the provided nodes and edges descriptions.

        :param node_descs: A list of `NodeDesc` objects.
        :param edge_descs: A list of `EdgeDesc` objects.
        :return: An instance of the `NNCFGraph` class.
        """
        nncf_graph = NNCFGraph()
        op_name_to_node_id_map = {}
        for desc in node_descs:
            nncf_node = nncf_graph.add_nncf_node(
                node_name=desc.op_name,
                node_type=desc.op_type_name,
                node_metatype=desc.metatype,
                layer_attributes=desc.attrs,
                layer_name=desc.resource_name if desc.resource_name else desc.op_name,
                is_shared=desc.is_shared,
            )
            op_name_to_node_id_map[desc.op_name] = nncf_node.node_id

        for desc in edge_descs:
            from_node_id = op_name_to_node_id_map[desc.producer_op_name]
            to_node_id = op_name_to_node_id_map[desc.consumer_op_name]

            nncf_graph.add_edge_between_nncf_nodes(
                from_node_id,
                to_node_id,
                tensor_shape=desc.tensor_shape,
                input_port_id=desc.input_port_id,
                output_port_id=desc.output_port_id,
                dtype=desc.tensor_dtype,
            )

        return nncf_graph

    @staticmethod
    def _get_data_format(op: tf.Operation) -> str:
        """
        Returns data format for the TensorFlow operation in the Keras format.
        Returns default data format if the operation does not have it.

        :param op: TensorFlow operation.
        :return: String `channels_last` or `channels_first`.
        """
        try:
            data_format = op.get_attr("data_format")
        except ValueError:
            data_format = None

        if data_format:
            to_keras_data_format = {
                "NHWC": "channels_last",
                "NCHW": "channels_first",
                "NDHWC": "channels_last",
                "NCDHW": "channels_first",
            }
            return to_keras_data_format[data_format.decode("utf-8")]

        return tf.keras.backend.image_data_format()

    @staticmethod
    def _collect_tfgraph_descs(graph: tf.Graph, op_names: List[str]) -> Tuple[List[NodeDesc], List[EdgeDesc]]:
        """
        Traverses the TF graph and collects information about nodes and edges
        which should be included to the NNCF graph.

        :param graph: Frozen `tf.Graph` with applied `dependency` optimization.
        :param op_names: A list of names for the input operations ()
        :return: A description of nodes and edges which should be included to the NNCF graph.
        """

        # Traverse the `graph` and mark all ops reachable from the `input_ops`.
        # The op `u` is reachable from the `input_ops` if a directed path
        # from at least one op in `input_ops` to `u` exists in the `graph.`

        # The NNCF graph contains only reachable nodes (ops).

        # If `op_name` in `visited_ops` then operation `visited_ops[op_name]`
        # (i.e. operation with `op_name` name) is reachable from the `input_ops`.
        input_ops: List[tf.Operation] = [graph.get_operation_by_name(name) for name in op_names]
        queue = deque(input_ops)
        visited_ops: Dict[str, tf.Operation] = {op.name: op for op in input_ops}
        while len(queue) != 0:
            v = queue.popleft()
            # A successor of node `v` is a node `u` such that exists
            # a directed edge (v, u) from `v` to `u`.
            successors = (op for tensor in v.outputs for op in tensor.consumers())
            for u in successors:
                if u.name not in visited_ops:
                    queue.append(u)
                    visited_ops[u.name] = u

        op_name_to_const_op_names_map = SubclassedConverter._get_op_name_to_const_op_names_map(graph, visited_ops)

        # Collect descriptions for all visited ops. The directed edge (v, u)
        # of the `graph` is added only if ops v, u were visited.
        node_descs = []
        edge_descs = []
        for op in visited_ops.values():
            metatype = get_op_metatype(op.type)
            node_attributes = TFNodeAttributes(SubclassedConverter._get_data_format(op))

            is_shared = False
            const_op_name = None
            if metatype in ALL_LAYER_METATYPES_WITH_WEIGHTS:
                const_op_names = op_name_to_const_op_names_map[op.name]
                if const_op_names:
                    # TODO(andrey-churkin): Currently, we don't have metatypes with
                    # multiple weights definitions.
                    if len(const_op_names) > 1 or len(metatype.weight_definitions) > 1:
                        raise NotImplementedError
                    const_op_name, is_shared = const_op_names[0]

                    # TODO(andrey-churkin): Seems like we can dynamically collect
                    # this information. Need to look at this.
                    port_id = metatype.weight_definitions[0].port_id
                    weight_shape = op.inputs[port_id].shape.as_list()

                    node_attributes = TFWeightedNodeAttributes(node_attributes.get_data_format(), weight_shape)

            node_descs.append(
                NodeDesc(
                    op_name=op.name,
                    op_type_name=op.type,
                    metatype=metatype,
                    is_shared=is_shared,
                    resource_name=const_op_name,
                    attrs=node_attributes,
                )
            )

            for input_port_id, tensor in enumerate(op.inputs):
                producer_op = tensor.op
                if producer_op.name not in visited_ops:
                    continue

                edge_descs.append(
                    EdgeDesc(
                        producer_op_name=producer_op.name,
                        output_port_id=tensor.value_index,
                        consumer_op_name=op.name,
                        input_port_id=input_port_id,
                        tensor_shape=tensor.shape.as_list(),
                        tensor_dtype=SubclassedConverter._convert_dtype_to_nncf_format(tensor.dtype),
                    )
                )

        return node_descs, edge_descs

    @staticmethod
    def _convert_dtype_to_nncf_format(dtype: tf.dtypes.DType) -> Dtype:
        """
        Converts `tf.dtypes.DType` to the NNCF representation.

        :param dtype: An instance of the `tf.dtypes.DType`.
        :return: An instance of the `Dtype`.
        """
        if dtype.is_floating:
            tensor_dtype = Dtype.FLOAT
        elif dtype.is_integer:
            tensor_dtype = Dtype.INTEGER
        else:
            raise nncf.InternalError(f"Unexpected dtype of tensor: {dtype}")

        return tensor_dtype

    @staticmethod
    def _get_op_name_to_const_op_names_map(
        graph, marked_ops: Dict[str, tf.Operation]
    ) -> Dict[str, List[Tuple[str, bool]]]:
        """
        Returns information about constant operations for the `marked_ops`.

        :param graph: Frozen tf.Graph.
        :param marked_ops: A mapping from operation name to operation. The marked operations are
            the operations which reachable from the input operations (`input_ops`) in the `graph`.
        :return: A mapping from the operation name to the list of (const_op_name, is_shared) tuples where
            - `const_op_name` is the name of constant operation which is used by this operation
            - `is_shared` is the boolean flag. Takes one of the following values:
            `True` if the number of operations that use it is greater than 1, `False` otherwise.
        """
        const_ops = [op for op in graph.get_operations() if op.type == "Const"]

        const_op_name_to_op_names_map: Dict[str, List[str]] = {}
        for const_op in const_ops:
            # Traverse the `graph` from the `const_op` and find the reachable ops
            # which are marked as visited (i.e. it's name in `marked_ops` dict)
            # from the `input_ops` in the previous traverse.
            queue = deque([const_op])
            visited_ops = {const_op.name: const_op}
            while len(queue) != 0:
                v = queue.popleft()
                if v.name in marked_ops:
                    const_op_name_to_op_names_map.setdefault(const_op.name, []).append(v.name)
                    continue
                # A successor of node `v` is a node `u` such that exists
                # a directed edge (v, u) from `v` to `u`.
                successors = (op for tensor in v.outputs for op in tensor.consumers())
                for u in successors:
                    if u.name not in visited_ops:
                        queue.append(u)
                        visited_ops[u.name] = u

        op_name_to_const_op_names_map: Dict[str, List[Tuple[str, bool]]] = {}
        for const_op_name, op_names in const_op_name_to_op_names_map.items():
            is_shared = len(op_names) > 1
            for op_name in op_names:
                op_name_to_const_op_names_map.setdefault(op_name, []).append((const_op_name, is_shared))

        return op_name_to_const_op_names_map
